from pathlib import Path
from typing import Iterable, Optional

import pandas as pd

__all__ = ["read_clinical_data", "read_imaging_data", "dataset_to_bids", "write_bids"]


def _find_clinical_data(clinical_data_directory: Path) -> Optional[pd.DataFrame]:
    dataframe = None
    for file in clinical_data_directory.rglob("*.xlsx"):
        try:
            dataframe = (
                pd.read_excel(file, index_col=[0, 4])
                .rename(columns=lambda x: x.lower().replace(" ", "_"))
                .rename_axis(index=lambda x: x.lower().replace(" ", "_"))
                .convert_dtypes()
                .sort_index()
            )
        except (ValueError, IndexError):
            continue

    return dataframe


def read_clinical_data(clinical_data_directory: Path) -> pd.DataFrame:
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    if (dataframe := _find_clinical_data(clinical_data_directory)) is None:
        raise FileNotFoundError("Clinical data not found")
    # Compute participant and session IDs.
    dataframe = dataframe.rename_axis(
        index={"loni_id": "participant_id", "visit_number": "session_id"}
    )
    dataframe.index = dataframe.index.map(
        lambda x: (
            bids_id_factory(StudyName.NIFD).from_original_study_id(x[0]),
            f"ses-M{(6 * (x[1] - 1)):03d}",
        )
    )

    # Keep relevant columns and rename them.
    dataframe = (
        dataframe[["dx", "site", "education", "race", "cdr_box_score", "mmse_tot"]]
        .rename(columns={"dx": "diagnosis", "cdr_box_score": "cdr", "mmse_tot": "mmse"})
        .astype(
            dtype={
                "diagnosis": pd.CategoricalDtype(
                    ["BV", "CON", "L_SD", "PATIENT (OTHER)", "PNFA", "SV"]
                ),
                "site": pd.CategoricalDtype(["UCSF", "MAYO", "MGH"]),
                "education": pd.Int64Dtype(),
                "race": pd.Int64Dtype(),
                "cdr": pd.Float64Dtype(),
                "mmse": pd.Float64Dtype(),
            }
        )
        .replace({"education": {99: pd.NA}, "race": {50: pd.NA, 99: pd.NA}})
    )
    # Keep positive MMSE values only.
    dataframe.mmse = dataframe.mmse.mask(dataframe.mmse < 0)

    return dataframe


def _find_collection_data(imaging_data_directory: Path) -> Optional[pd.DataFrame]:
    dataframe = None
    for file in imaging_data_directory.rglob("*.csv"):
        try:
            dataframe = (
                pd.read_csv(
                    file,
                    index_col="Image Data ID",
                    parse_dates=["Acq Date"],
                )
                .rename(columns=lambda x: x.lower().replace(" ", "_"))
                .rename_axis(index=lambda x: x.lower().replace(" ", "_"))
                .convert_dtypes()
                .sort_index()
            )
        except (ValueError, IndexError):
            continue

    return dataframe


def _find_imaging_data(imaging_data_directory: Path) -> Iterable[tuple[str, str]]:
    for image_data_id, source_dir in set(
        sorted(_extract_id_with_dir(_find_files(imaging_data_directory)))
    ):
        yield image_data_id, source_dir


def _extract_id_with_dir(files: Iterable[Path]) -> tuple[str, str]:
    """Extract the image data ID from the NIFD files."""
    import re

    pattern = re.compile(r"(I\d{6})")
    for file in files:
        if (found := pattern.search(file.name)) is not None:
            yield found.group(1), str(file.parent)


def _find_files(in_: Path) -> Iterable[Path]:
    return filter(lambda x: x.is_file(), Path(in_).rglob("NIFD*.*"))


def read_imaging_data(imaging_data_directory: Path) -> pd.DataFrame:
    try:
        imaging_data = pd.DataFrame.from_records(
            data=_find_imaging_data(imaging_data_directory),
            columns=["image_data_id", "source_dir"],
            index="image_data_id",
        ).convert_dtypes()
    except TypeError:
        raise FileNotFoundError("No imaging data found")
    if (collection_data := _find_collection_data(imaging_data_directory)) is None:
        raise FileNotFoundError("No collection data found")

    return collection_data.join(imaging_data)


def _parse_mri_description(description: str) -> Optional[dict[str, Optional[str]]]:
    description = description.lower().replace("-", "")

    if (suffix := _infer_bids_suffix_from_description(description)) is None:
        return None
    return {
        "datatype": "anat",
        "suffix": suffix,
        "trc_label": None,
        "rec_label": None,
    }


def _infer_bids_suffix_from_description(description: str) -> Optional[str]:
    if any([x in description for x in ("mprage", "mt1", "gradwarp", "n3m")]):
        return "T1w"
    if "flair" in description:
        return "FLAIR"
    if "t2" in description:
        return "T2w"
    if "asl" in description:
        return "PDw"
    return None


def _parse_pet_description(description: str) -> Optional[dict[str, str]]:
    import re

    match = re.search(r"3D:(\w+):(\w+)", description)

    if match:
        return {
            "datatype": "pet",
            "suffix": "pet",
            "trc_label": "11CPIB" if "PIB" in match.group(1) else "18FFDG",
            "rec_label": "IR" if "IR" in match.group(2) else "RP",
        }
    return None


def _parse_preprocessing(description: str) -> dict[str, bool]:
    description = description.lower()

    return {
        "gradwarp": any([x in description for x in ["gradwarp", "dis3d"]]),
        "n3": "n3m" in description,
    }


def dataset_to_bids(
    imaging_data: pd.DataFrame,
    clinical_data: Optional[pd.DataFrame] = None,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    # Parse preprocessing information from scan descriptions.
    preprocessing = imaging_data.description.apply(_parse_preprocessing).apply(
        pd.Series
    )
    # Parse BIDS entities from scan descriptions.
    bids = (
        imaging_data.apply(
            lambda x: _parse_pet_description(x.description)
            if x.modality == "PET"
            else _parse_mri_description(x.description),
            axis=1,
        )
        .dropna()
        .apply(pd.Series)
    )
    # Compute quality metric for each scan:
    # - MRI: Applied preprocessing (0: None, 1: GradWarp, 2: N3)
    # - PET: Reconstruction method (0: Fourier, 1: Iterative)
    quality = (
        preprocessing.sum(axis=1)
        + bids.rec_label.apply(lambda x: 1 if x == "IR" else 0)
    ).rename("quality")
    # Select one scan per BIDS modality based on quality metric.
    subset = ["subject", "visit", "datatype", "suffix", "trc_label"]
    scans = (
        bids.join(quality)
        .join(imaging_data)
        .sort_values(by=subset + ["quality"])
        .drop_duplicates(subset=subset, keep="last")
        .drop(columns="quality")
    )
    # Compute the BIDS-compliant participant, session and scan IDs.
    scans = scans.assign(
        participant_id=lambda df: df.subject.apply(
            lambda x: bids_id_factory(StudyName.NIFD).from_original_study_id(x)
        ),
        session_id=lambda df: df.visit.apply(
            lambda x: (
                "ses-M" + x.strip("M").zfill(3)
                if isinstance(x, str)
                else f"ses-M{(6 * (x - 1)):03d}"
            )
        ),
        filename=lambda df: df.apply(
            lambda x: (
                f"{x.participant_id}/{x.session_id}/{x.datatype}/"
                f"{x.participant_id}_{x.session_id}"
                f"{'_trc-' + x.trc_label if x.trc_label else ''}"
                f"{'_rec-' + x.rec_label if x.rec_label else ''}"
                f"_{x.suffix}.nii.gz"
            ),
            axis=1,
        ),
    )
    # Prepare subjects manifest.
    subjects = (
        scans[["participant_id", "session_id", "group", "sex", "age"]]
        .sort_values(by=["participant_id", "session_id"])
        .drop(columns="session_id")
        .drop_duplicates(subset="participant_id")
        .set_index(["participant_id"], verify_integrity=True)
        .sort_index()
    ).join(
        clinical_data.xs("ses-M000", level="session_id")[
            ["diagnosis", "site", "education", "race"]
        ]
    )
    # Prepare sessions manifest.
    sessions = (
        scans[["participant_id", "session_id", "acq_date", "age"]]
        .rename(columns={"acq_date": "date"})
        .drop_duplicates()
        .set_index(["participant_id", "session_id"], verify_integrity=True)
        .sort_index()
    ).join(clinical_data[["cdr", "mmse"]])
    # Prepare scans manifest.
    scans = scans[["filename", "source_dir", "format"]].set_index(
        "filename", verify_integrity=True
    )

    return subjects, sessions, scans


def _convert_dicom(sourcedata_dir: Path, bids_filename: Path) -> None:
    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_utils import run_dcm2niix

    output_fmt = str(bids_filename.name).replace(".nii.gz", "")
    output_dir = bids_filename.parent

    # Ensure output directory is empty.
    fs = LocalFileSystem()
    if fs.exists(output_dir):
        fs.rm(str(output_dir), recursive=True)
    fs.makedirs(str(output_dir))

    # Run conversion with dcm2niix with anonymization and maximum compression.
    run_dcm2niix(
        input_dir=sourcedata_dir,
        output_dir=output_dir,
        output_fmt=output_fmt,
        compress=True,
        bids_sidecar=True,
    )


def _install_nifti(sourcedata_dir: Path, bids_filename: Path) -> None:
    from fsspec.implementations.local import LocalFileSystem

    fs = LocalFileSystem(auto_mkdir=True)
    source_file = fs.open(fs.ls(str(sourcedata_dir))[0], mode="rb")
    target_file = fs.open(str(bids_filename), mode="wb", compression="gzip")

    with source_file as sf, target_file as tf:
        tf.write(sf.read())


def write_bids(
    to: Path,
    participants: pd.DataFrame,
    sessions: pd.DataFrame,
    scans: pd.DataFrame,
) -> list[Path]:
    from fsspec.implementations.local import LocalFileSystem

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import StudyName, write_to_tsv

    fs = LocalFileSystem(auto_mkdir=True)

    # Ensure BIDS hierarchy is written first.
    with fs.transaction:
        with fs.open(
            str(to / "dataset_description.json"), "w"
        ) as dataset_description_file:
            BIDSDatasetDescription(name=StudyName.NIFD).write(
                to=dataset_description_file
            )
        with fs.open(str(to / "participants.tsv"), "w") as participant_file:
            write_to_tsv(participants, participant_file)

        for participant_id, sessions_group in sessions.groupby("participant_id"):
            participant_id = str(participant_id)
            sessions_group = sessions_group.droplevel("participant_id")
            sessions_filepath = to / participant_id / f"{participant_id}_sessions.tsv"
            with fs.open(str(sessions_filepath), "w") as sessions_file:
                write_to_tsv(sessions_group, sessions_file)

    for filename, metadata in scans.iterrows():
        filename = str(filename)
        (_convert_dicom if metadata.format == "DCM" else _install_nifti)(
            sourcedata_dir=metadata.source_dir, bids_filename=to / filename
        )

    return scans.index.to_list()



from typing import List

from pydra.mark import annotate, task


@task
@annotate({"return": {"nii_file": str}})
def peak_correction_task(
    t_map: str, t_threshold: float, output_name: str = None
) -> str:
    """Pydra task for correcting the peak.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils.peak_correction`.
    """
    from clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils import (
        peak_correction,
    )

    return peak_correction(t_map, t_threshold, output_name)


@task
@annotate({"return": {"nii_file": str}})
def cluster_correction_task(
    t_map: str, t_thresh: float, c_thresh: int, output_name: str = None
) -> str:
    """Pydra task for cluster correction

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils.cluster_correction`.
    """
    from clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils import (
        cluster_correction,
    )

    return cluster_correction(t_map, t_thresh, c_thresh, output_name)


@task
@annotate({"return": {"figs": list}})
def produce_figures_task(
    nii_file: str,
    template: str,
    type_of_correction: str,
    t_thresh: str,
    c_thresh: int,
    n_cuts: int,
) -> List[str]:
    """Pydra task for figure production

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils.produce_figures`
    """
    from clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils import (
        produce_figures,
    )

    return produce_figures(
        nii_file, template, type_of_correction, t_thresh, c_thresh, n_cuts
    )


@task
@annotate({"return": {"bool": bool}})
def generate_output_task(t_map: str, figs: List[str], correction_name: str) -> None:
    """Pydra task for output production

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils.generate_output`
    """
    from clinica.pipelines.statistics_volume_correction.statistics_volume_correction_utils import (
        generate_output,
    )

    return generate_output(t_map, figs, correction_name)



from .parsers import parse_bids_file
from .readers import read_bids, read_bids_dataset, read_bids_files
from .templates import (
    download_mni_template_2009a,
    download_mni_template_2009c,
    download_ref_template,
)
from .writers import write_bids_file, write_caps_file



import typing as ty

import pydra


@pydra.mark.task
@pydra.mark.annotate({"return": {"subject_id": "str"}})
def compute_freesurfer_subject_id(participant_id: str, session_id: ty.Optional[str]):
    """Compute FreeSurfer's subject ID from BIDS.

    Unlike BIDS, FreeSurfer's layout does not encode the concept of sessions.

    The latter is emulated using a compound subject ID, like `sub-P01_ses-M00`
    for BIDS participant `sub-P01` at imaging session `ses-M00`.

    If the BIDS dataset is single-session,
    then use the participant ID as FreeSurfer's subject ID.

    Parameters
    ----------
    participant_id : str
        BIDS subject identifier.
    session_id : str, optional
        BIDS session identifier.

    Returns
    -------
    str
        FreeSurfer's subject identifier.
    """
    return f"{participant_id}_{session_id}" if session_id else participant_id



from ._plot import StatisticsResultsPlotter
from ._serialize import StatisticsResultsSerializer
from ._statistics import StatisticsResults

__all__ = [
    "StatisticsResults",
    "StatisticsResultsSerializer",
    "StatisticsResultsPlotter",
]



from dataclasses import dataclass
from enum import Enum
from os import PathLike
from pathlib import Path
from typing import Callable, Optional, Tuple, Union

import nibabel as nib
import numpy as np
from nibabel.nifti1 import Nifti1Image

__all__ = [
    "HemiSphere",
    "compute_aggregated_volume",
    "get_new_image_like",
    "merge_nifti_images_in_time_dimension",
    "remove_dummy_dimension_from_image",
    "crop_nifti",
    "get_mni_template",
    "get_mni_cropped_template",
]


class HemiSphere(str, Enum):
    """Possible values for hemispheres."""

    LEFT = "lh"
    RIGHT = "rh"


def compute_aggregated_volume(
    image_filename: PathLike,
    aggregator: Optional[Callable] = None,
    volumes_to_keep: Optional[np.ndarray] = None,
) -> np.ndarray:
    """Computes the aggregated 3D volumes from a 4D image and an aggregator function.

    The aggregation is computed on the last (fourth) dimension.
    It is possible to compute the aggregation on a subset of the volumes through the
    parameter 'volumes_to_keep'.

    Parameters
    ----------
    image_filename : str
        The path to the input image.

    aggregator : Callable, optional
        The aggregator function. Example: np.average, np.median.
        If None, no aggregation will be performed.

    volumes_to_keep : np.array, optional
        The volumes to be kept during aggregation. This is a 1D index array.
        If None, all volumes of the input image will be kept during aggregation.
        Default = None.

    Returns
    -------
    np.ndarray:
        The 3D volume data array obtained from the aggregation.
    """
    volumes = np.array(nib.four_to_three(nib.load(image_filename)))
    if volumes_to_keep is None:
        volumes_to_keep = volumes
    else:
        volumes_to_keep = volumes[volumes_to_keep]
    data = [
        volume.get_fdata(dtype="float32").astype(np.float32)
        for volume in volumes_to_keep
    ]
    if aggregator is None:
        return np.stack(data, axis=-1)
    return aggregator(data, axis=0)


def get_new_image_like(old_image: PathLike, new_image_data: np.ndarray) -> Nifti1Image:
    """Builds a new Nifti1Image from the provided image and new data.

    Parameters
    ----------
    old_image : PathLike
        The path to the old image file from which to get the header and affine.

    new_image_data : np.ndarray
        The data for the new image to build.

    Returns
    -------
    Nifti1Image :
        The new image.
    """
    old_img = nib.load(old_image)
    hdr = old_img.header.copy()
    hdr.set_data_shape(new_image_data.shape)
    hdr.set_xyzt_units("mm")
    hdr.set_data_dtype(np.float32)

    return nib.Nifti1Image(new_image_data, old_img.affine, hdr)


def merge_nifti_images_in_time_dimension(
    images: Tuple[Union[str, PathLike], ...], out_file: Optional[PathLike] = None
) -> PathLike:
    """Concatenates the provided images in the 4th dimension.

    The provided images must all be 3D or 4D. For 3D images, a dummy
    4th dimension will be added before concatenation.

    Parameters
    ----------
    images : tuple of str or tuple of Pathlike
        The paths to the images that should get merged.

    out_file : PathLike, optional
        The path to the file in which to write the merged
        volumes. If None, the volumes will be written to a
        file named 'merged_files.nii.gz' in the current folder.

    Returns
    -------
    out_file : PathLike
        Path to merged volumes.
    """
    import os

    images = _check_existence(images)
    out_file = out_file or os.path.abspath("merged_files.nii.gz")
    volumes = _check_volumes_from_images(images)
    merged_volume = np.concatenate(volumes, axis=-1)
    merged_image = get_new_image_like(images[0], merged_volume)
    nib.save(merged_image, out_file)

    return out_file


def _check_existence(filenames: Tuple[PathLike, ...]) -> Tuple[Path, ...]:
    """Converts all element in provided tuple to Path objects and check for existence."""
    from pathlib import Path

    if len(filenames) < 2:
        raise ValueError("At least 2 files are required.")
    filenames = tuple(Path(f) for f in filenames)
    missing_files = tuple(f for f in filenames if not f.exists())
    if missing_files:
        raise FileNotFoundError(f"the following file(s) are missing: {missing_files}")
    return filenames


def _check_volumes_from_images(images: Tuple[Path, ...]) -> Tuple[np.ndarray, ...]:
    """Loads the images and check the dimensions."""
    images = tuple(nib.load(i) for i in images)
    volumes = tuple(i.get_fdata() for i in images)
    four_dimensional_volumes = []
    for volume in volumes:
        if volume.ndim == 3:
            four_dimensional_volumes.append(volume[..., np.newaxis])
        elif volume.ndim == 4:
            four_dimensional_volumes.append(volume)
        else:
            raise ValueError(
                f"Only 3D or 4D images can be concatenated. A {volume.ndim}D image was found."
            )

    return tuple(four_dimensional_volumes)


def remove_dummy_dimension_from_image(image: str, output: str) -> str:
    """Remove all dummy dimensions (i.e. equal to 1) from an image.

    Parameters
    ----------
    image : str
        Path to the input image.

    output : str
        Path to the desired output image.

    Returns
    -------
    str :
        The path to the output image.
    """
    import nibabel as nib
    from nilearn.image import new_img_like

    img = new_img_like(image, nib.load(image).get_fdata().squeeze())
    nib.save(img, output)

    return output


@dataclass
class Slice:
    """Interval composed of a starting point and ending point."""

    start: int
    end: int

    def __post_init__(self):
        if self.end < self.start:
            raise ValueError(
                f"Slice instance has a start value ({self.start}) larger than the end value ({self.end})."
            )

    def get_slice(self) -> slice:
        return slice(self.start, self.end)

    def __repr__(self) -> str:
        return f"( {self.start}, {self.end} )"


@dataclass
class Bbox3D:
    """3D Bounding Box."""

    x_slice: Slice
    y_slice: Slice
    z_slice: Slice

    @classmethod
    def from_coordinates(
        cls,
        start_x: int,
        end_x: int,
        start_y: int,
        end_y: int,
        start_z: int,
        end_z: int,
    ):
        return cls(
            Slice(start_x, end_x),
            Slice(start_y, end_y),
            Slice(start_z, end_z),
        )

    def get_slices(self) -> Tuple[slice, slice, slice]:
        return (
            self.x_slice.get_slice(),
            self.y_slice.get_slice(),
            self.z_slice.get_slice(),
        )

    def __repr__(self):
        return f"( {self.x_slice}, {self.y_slice}, {self.z_slice} )"


# This bounding box has been pre-computed by clinica developers
MNI_CROP_BBOX = Bbox3D.from_coordinates(12, 181, 13, 221, 0, 179)


def _crop_array(array: np.ndarray, bbox: Bbox3D) -> np.ndarray:
    # TODO: When Python 3.10 is dropped, replace with 'return array[*bbox.get_slices()]'
    x, y, z = bbox.get_slices()
    return array[x, y, z]


def _get_file_locally_or_download(
    filename: str, url: Optional[str] = None, expected_checksum: Optional[str] = None
) -> Path:
    from clinica.utils.inputs import RemoteFileStructure, fetch_file

    resource_folder = Path(__file__).parent.parent / "resources" / "masks"
    local_file = resource_folder / filename
    if not local_file.exists():
        if url is None or expected_checksum is None:
            raise ValueError(
                f"Downloading file {filename} requires both the URL and the expected checksum "
                "such that it can be ensured that no data corruption occurred."
            )
        fetch_file(
            RemoteFileStructure(
                filename=filename,
                url=url,
                checksum=expected_checksum,
            ),
            resource_folder,
        )
    if local_file.exists():
        return local_file
    raise FileNotFoundError(f"Unable to get file {filename} locally or remotely.")


def get_mni_cropped_template() -> Path:
    return _get_file_locally_or_download(
        filename="ref_cropped_template.nii.gz",
        url="https://aramislab.paris.inria.fr/files/data/img_t1_linear/",
        expected_checksum="67e1e7861805a8fd35f7fcf2bdf9d2a39d7bcb2fd5a201016c4d2acdd715f5b3",
    )


def get_mni_template(modality: str) -> Path:
    """Get the path to the MNI template for the given modality.

    If the file can be found locally in the resources folder, it is
    returned directly, otherwise it is downloaded from the aramislab
    server.

    Parameters
    ----------
    modality : str
        t1 or flair depending on which template is desired.

    Returns
    -------
    Path :
        The path to the required MNI template.

    Raises
    ------
    ValueError:
        If the modality is not t1 or flair.
    FileNotFoundError:
        If the template could not be retrieved locally or remotely.
    """
    if modality.lower() == "t1":
        return _get_mni_template_t1()
    if modality.lower() == "flair":
        return _get_mni_template_flair()
    raise ValueError(f"No MNI template available for modality {modality}.")


def _get_mni_template_t1() -> Path:
    return _get_file_locally_or_download(
        filename="mni_icbm152_t1_tal_nlin_sym_09c.nii",
        url="https://aramislab.paris.inria.fr/files/data/img_t1_linear/",
        expected_checksum="93359ab97c1c027376397612a9b6c30e95406c15bf8695bd4a8efcb2064eaa34",
    )


def _get_mni_template_flair() -> Path:
    return _get_file_locally_or_download(
        filename="GG-853-FLAIR-1.0mm.nii.gz",
        url="https://aramislab.paris.inria.fr/files/data/img_flair_linear/",
        expected_checksum="b1d2d359a4c3671685227bb14014ce50ac232012b628335a4c049e2911c64ce1",
    )


def crop_nifti(input_image: Path, output_dir: Optional[Path] = None) -> Path:
    """Crop input image.

    The function expects a 3D anatomical image and will crop it
    using a pre-computed bounding box. This bounding box has been
    used to crop the MNI template (located in
    clinica/resources/masks/mni_icbm152_t1_tal_nlin_sym_09c.nii)
    into the cropped template (located in
    clinica/resources/masks/ref_cropped_template.nii.gz).

    Parameters
    ----------
    input_image : Path
        The path to the input image to be cropped.

    output_dir : Path, optional
        The folder in which to write the output cropped image.
        If not provided, the image will be written in current folder.

    Returns
    -------
    output_img : Path
        The path to the cropped image.

    Raises
    ------
    ValueError:
        If the input image is not 3D.
    """
    from nilearn.image import new_img_like

    from clinica.utils.filemanip import get_filename_no_ext

    filename_no_ext = get_filename_no_ext(input_image)
    input_image = nib.load(input_image)
    if len(input_image.shape) != 3:
        raise ValueError(
            "The function crop_nifti is implemented for anatomical 3D images. "
            f"You provided an image of shape {input_image.shape}."
        )
    output_dir = output_dir or Path.cwd()
    crop_img = new_img_like(
        nib.load(get_mni_cropped_template()),
        _crop_array(input_image.get_fdata(), MNI_CROP_BBOX),
    )
    output_img = output_dir / f"{filename_no_ext}_cropped.nii.gz"
    crop_img.to_filename(output_img)

    return output_img



__all__ = [
    "init_input_node_task",
    "move_subjects_dir_to_source_dir_task",
    "save_to_caps_task",
    "write_tsv_files_task",
]


def save_to_caps_task(
    source_dir: str, subject_id: str, caps_dir: str, overwrite_caps: bool = False
) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.correction.utils import (
        save_to_caps,
    )

    return save_to_caps(
        Path(source_dir),
        subject_id,
        Path(caps_dir),
        overwrite_caps=overwrite_caps,
    )


def move_subjects_dir_to_source_dir_task(
    subjects_dir: str, source_dir: str, subject_id: str
) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.utils import (
        move_subjects_dir_to_source_dir,
    )
    from clinica.pipelines.anatomical.freesurfer.utils import (
        extract_image_id_from_freesurfer_id,
    )

    return move_subjects_dir_to_source_dir(
        Path(subjects_dir),
        Path(source_dir),
        subject_id,
        image_id="_".join(extract_image_id_from_freesurfer_id(subject_id)),
    )


def init_input_node_task(
    caps_dir: str, participant_id: str, session_id: str, long_id: str, output_dir: str
) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.correction.utils import (
        init_input_node,
    )

    return str(
        init_input_node(
            Path(caps_dir),
            participant_id,
            session_id,
            long_id,
            Path(output_dir),
        )
    )


def write_tsv_files_task(subjects_dir: str, subject_id: str) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.longitudinal.correction.utils import (
        write_tsv_files,
    )

    return write_tsv_files(Path(subjects_dir), subject_id)



import pydra
from pydra import Workflow

from clinica.pydra.engine import clinica_io


@clinica_io
def build_core_workflow(name: str = "core", parameters={}) -> Workflow:
    """Build the core workflow for the Statistics Volume pipeline.

    Parameters
    ----------
    name : str, optional
        The name of the workflow. Default="core".

    parameters : dict, optional
        Dictionary of parameters to be used
        within the workflow.
        Default={}.

    Returns
    -------
    wf : Workflow
        The core workflow.
    """
    from typing import Any

    import numpy as np

    import clinica.pydra.statistics_volume_correction.task as utils
    from clinica.pydra.tasks import download_mni_template_2009a
    from clinica.utils.spm import use_spm_standalone_if_available

    use_spm_standalone_if_available()

    query = {"pattern": parameters["t_map"] + "*", "description": "statistics t map"}

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            ("t_map", dict, query, {"mandatory": True}),
        ],
        bases=(pydra.specs.BaseSpec,),
    )
    wf = Workflow(name, input_spec=input_spec)

    for threshold in ("FWE", "FDR"):
        wf.add(
            utils.peak_correction_task(
                name=f"{threshold}_peak_correction_task",
                t_map=wf.lzin.t_map,
                t_threshold=parameters[f"{threshold}p"],
            )
        )
    for threshold in ("FWE", "FDR"):
        wf.add(
            utils.cluster_correction_task(
                name=f"{threshold}_cluster_correction_task",
                t_map=wf.lzin.t_map,
                t_thresh=parameters["height_threshold"],
                c_thresh=parameters[f"{threshold}c"],
            )
        )

    wf.add(download_mni_template_2009a(name="download_mni_template"))

    for threshold in ("FWE", "FDR"):
        for kind in ("peak", "cluster"):
            t_thresh_key = f"{threshold}p" if kind == "peak" else "height_threshold"
            c_thresh = parameters[f"{threshold}c"] if kind == "cluster" else np.nan
            wf.add(
                utils.produce_figures_task(
                    name=f"produce_figure_{threshold}_{kind}_correction",
                    nii_file=getattr(
                        wf, f"{threshold}_{kind}_correction_task"
                    ).lzout.nii_file,
                    template=wf.download_mni_template.lzout.mni_template_file,
                    type_of_correction=threshold,
                    t_thresh=parameters[t_thresh_key],
                    c_thresh=c_thresh,
                    n_cuts=parameters["n_cuts"],
                )
            )
            wf.add(
                utils.generate_output_task(
                    name=f"save_figure_{kind}_correction_{threshold}",
                    t_map=wf.lzin.t_map,
                    figs=getattr(
                        wf, f"produce_figure_{threshold}_{kind}_correction"
                    ).lzout.figs,
                    correction_name=f"{threshold}{kind[0]}",
                )
            )
    wf.set_output([("figs", wf.produce_figure_FDR_peak_correction.lzout.figs)])
    return wf



import json
import random
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Optional

import nibabel as nib
import numpy as np
from numpy.testing import assert_array_almost_equal, assert_array_equal

from clinica.pipelines.dwi.utils import DWIDataset

__all__ = [
    "build_bids_directory",
    "build_caps_directory",
    "build_dwi_dataset",
    "rmtree",
    "assert_nifti_equal",
    "assert_nifti_almost_equal",
    "assert_large_nifti_almost_equal",
]


def build_bids_directory(
    directory: Path,
    subjects_sessions: dict,
    modalities: Optional[dict[str, tuple[str, ...]]] = None,
    write_tsv_files: bool = False,
    random_seed: int = 42,
) -> Path:
    """Build a fake BIDS dataset at the specified location following the
    specified structure.

    Parameters
    ----------
    directory : Path
        The path to folder where the fake BIDS dataset should be created.

    subjects_sessions : Dict
        Dictionary containing the subjects and their associated sessions.

    modalities : dict
        Modalities to be created.

    Returns
    -------
    Path :
        The path to the BIDS dataset.

    Notes
    -----
    This function is a simple prototype for creating fake datasets for testing.
    It only adds (for now...) T1W nifti images for all subjects and sessions.
    """
    from clinica.utils.bids import BIDS_VERSION

    random.seed(random_seed)
    directory.mkdir(exist_ok=True, parents=True)
    modalities = modalities or {"anat": ("T1w", "flair")}
    extensions = ("nii.gz", "json")
    with open(directory / "dataset_description.json", "w") as fp:
        json.dump({"Name": "Example dataset", "BIDSVersion": str(BIDS_VERSION)}, fp)
    for sub, sessions in subjects_sessions.items():
        (directory / sub).mkdir()
        if write_tsv_files:
            (directory / sub / f"{sub}_sessions.tsv").write_text(
                "\n".join(["session_id"] + sessions)
            )
        for ses in sessions:
            session_path = directory / sub / ses
            session_path.mkdir()
            scans_tsv_text = "filename\tscan_id\n"
            for modality, suffixes in modalities.items():
                (session_path / modality).mkdir()
                for suffix in suffixes:
                    for extension in extensions:
                        scan_path = (
                            session_path
                            / modality
                            / f"{sub}_{ses}_{suffix}.{extension}"
                        )
                        scan_path.touch()
                        if write_tsv_files and "nii" in extension:
                            scans_tsv_text += f"{scan_path.relative_to(session_path)}\t{random.randint(1, 1000000)}\n"
            if write_tsv_files and len(scans_tsv_text) > 0:
                (session_path / f"{sub}_{ses}_scans.tsv").write_text(scans_tsv_text)
    return directory


def build_caps_directory(directory: Path, configuration: dict) -> Path:
    """Build a fake CAPS dataset at the specified location following the
    specified structure.

    Parameters
    ----------
    directory : Path
        The path to folder where the fake CAPS dataset should be created.

    configuration : Dict
        Dictionary containing the configuration for building the fake CAPS
        directory. It should have the following structure:
            - "groups": ["group_labels"...]
            - "pipelines": ["pipeline_names"...]
            - "subjects": {"subject_labels": ["session_labels"...]}.

    Returns
    -------
    Path :
        The path to the CAPS dataset.

    Notes
    -----
    This function is a simple prototype for creating fake datasets for testing.
    """
    from clinica.utils.bids import BIDS_VERSION
    from clinica.utils.caps import CAPS_VERSION

    directory.mkdir(exist_ok=True, parents=True)
    with open(directory / "dataset_description.json", "w") as fp:
        json.dump(
            {
                "Name": "Example dataset",
                "BIDSVersion": str(BIDS_VERSION),
                "CAPSVersion": str(CAPS_VERSION),
            },
            fp,
        )
    _build_groups(directory, configuration)
    _build_subjects(directory, configuration)
    return directory


def _build_groups(directory: Path, configuration: dict) -> None:
    """Build a fake groups file structure in a CAPS directory if requested."""
    if "groups" in configuration:
        (directory / "groups").mkdir()
        for group_label in configuration["groups"]:
            (directory / "groups" / f"group-{group_label}").mkdir()
            for pipeline in configuration["pipelines"]:
                (directory / "groups" / f"group-{group_label}" / pipeline).mkdir()
                (
                    directory
                    / "groups"
                    / f"group-{group_label}"
                    / pipeline
                    / f"group-{group_label}_template.nii.gz"
                ).touch()


def _build_subjects(directory: Path, configuration: dict) -> None:
    """Build a fake subjects file structure in a CAPS directory if requested."""
    if "subjects" in configuration:
        (directory / "subjects").mkdir()
        for sub, sessions in configuration["subjects"].items():
            (directory / "subjects" / sub).mkdir()
            for ses in sessions:
                (directory / "subjects" / sub / ses).mkdir()
                for pipeline in configuration["pipelines"]:
                    (directory / "subjects" / sub / ses / pipeline).mkdir()
                    if pipeline == "t1_linear":
                        _build_t1_linear(directory, sub, ses)
                    if pipeline == "t1":
                        _build_t1(directory, sub, ses, configuration)


def _build_t1_linear(directory: Path, sub: str, ses: str) -> None:
    """Build a fake t1-linear file structure in a CAPS directory."""
    (
        directory
        / "subjects"
        / sub
        / ses
        / "t1_linear"
        / f"{sub}_{ses}_T1w_space-MNI152NLin2009cSym_res-1x1x1_T1w.nii.gz"
    ).touch()


def _build_t1(directory: Path, sub: str, ses: str, configuration: dict) -> None:
    """Build a fake t1 file structure in a CAPS directory."""
    (directory / "subjects" / sub / ses / "t1" / "spm").mkdir()
    for folder in ["dartel", "segmentation"]:
        (directory / "subjects" / sub / ses / "t1" / "spm" / folder).mkdir()
    for group_label in configuration["groups"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "dartel"
            / f"group-{group_label}"
        ).mkdir()
    (
        directory
        / "subjects"
        / sub
        / ses
        / "t1"
        / "spm"
        / "dartel"
        / f"group-{group_label}"
        / f"{sub}_{ses}_T1w_target-{group_label}_transformation-forward_deformation.nii.gz"
    ).touch()
    for tissue in ["csf", "graymatter", "whitematter"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "dartel"
            / f"group-{group_label}"
            / f"{sub}_{ses}_T1w_segm-{tissue}_space-Ixi549Space_modulated-on_probability.nii.gz"
        ).touch()
    for folder in ["dartel_input", "native_space", "normalized_space"]:
        (
            directory / "subjects" / sub / ses / "t1" / "spm" / "segmentation" / folder
        ).mkdir()
    for tissue in ["csf", "graymatter", "whitematter"]:
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "dartel_input"
            / f"{sub}_{ses}_T1w_segm-{tissue}_dartelinput.nii.gz"
        ).touch()
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "native_space"
            / f"{sub}_{ses}_T1w_segm-{tissue}_probability.nii.gz"
        ).touch()
        (
            directory
            / "subjects"
            / sub
            / ses
            / "t1"
            / "spm"
            / "segmentation"
            / "normalized_space"
            / f"{sub}_{ses}_T1w_segm-{tissue}_space-Ixi549Space_modulated-off_probability.nii.gz"
        ).touch()


def rmtree(f: Path):
    if f.is_file():
        f.unlink()
    else:
        for child in f.iterdir():
            rmtree(child)
        f.rmdir()


def _assert_nifti_relation(
    img1: Path,
    img2: Path,
    assertion_func_data: Callable,
    assertion_func_affine: Optional[Callable] = None,
) -> None:
    """Assert that two nifti images satisfy some relationship.

    Parameters
    ----------
    img1 : Path
        Path to the first image.

    img2 : Path
        Path to the second image.

    assertion_func_data : Callable
        Assertion function for dataobj comparison.
        The function should take two numpy arrays as input.

    assertion_func_affine : Callable, optional
        Assertion function for affine comparison.
        The function should take two numpy arrays as input.
        If not specified, `assertion_func_data` will be used to
        check affine matrices.

    See Also
    --------
    assert_nifti_equal : Check strict equality for two images.
    assert_nifti_almost_equal : Check loose equality for two images.
    """
    assertion_func_affine = assertion_func_affine or assertion_func_data
    img1 = nib.load(img1)
    img2 = nib.load(img2)

    assert img1.shape == img2.shape  # Fail fast
    assertion_func_affine(img1, img2)
    assertion_func_data(img1, img2)


def _assert_affine_equal(img1: nib.Nifti1Image, img2: nib.Nifti1Image) -> None:
    assert_array_equal(img1.affine, img2.affine)


def _assert_dataobj_equal(img1: nib.Nifti1Image, img2: nib.Nifti1Image) -> None:
    assert_array_equal(img1.get_fdata(), img2.get_fdata())


def _assert_affine_almost_equal(
    img1: nib.Nifti1Image, img2: nib.Nifti1Image, decimal: int = 6
) -> None:
    assert_array_almost_equal(img1.affine, img2.affine, decimal=decimal)


def _assert_dataobj_almost_equal(
    img1: nib.Nifti1Image, img2: nib.Nifti1Image, decimal: int = 6
) -> None:
    assert_array_almost_equal(img1.get_fdata(), img2.get_fdata(), decimal=decimal)


def _assert_large_image_dataobj_almost_equal(
    img1: nib.Nifti1Image,
    img2: nib.Nifti1Image,
    decimal: int = 6,
    n_samples: Optional[int] = None,
    verbose: bool = False,
) -> None:
    """This alternative implementation can be used when dealing with large nifti images.

    Memory issue:

    `_assert_dataobj_almost_equal` loads both images data arrays in memory before running
    the equality assertion test which can be problematic when these arrays are large.
    This function will work over the last dimension (usually time, dwi directions...).
    It loops over this dimension and loads in memory only the two data chunks to be compared.

    CPU time issue:

    Even without the previously described memory issue, the comparison can take a lot
    of time for very large images. This function enables to sample volumes for equality
    assertion through the `n_samples` argument.

    TODO: Implement in parallel?

    Parameters
    ----------
    img1 : Nifti1Image
        The first image to be compared.

    img2 : Nifti1Image
        The second image to be compared.

    decimal : int, optional
        The number of decimal for which the loose equality should work.
        Default=6.

    n_samples : int, optional
        If specified, the function will only compare a subset of the data.
        If None, it will compare all data.
        The number of samples specifies the number of "volumes" which will be compared.
        Default=None.

    verbose : bool, optional
        If True, the function prints a message for each compared volume.
        Default=False.

    See Also
    --------
    assert_large_nifti_almost_equal : Check loose equality for two large images.
    """
    import random

    import numpy as np

    volumes = range(0, img1.shape[-1])
    if n_samples:
        volumes = random.sample(volumes, n_samples)
    for volume in volumes:
        if verbose:
            print(f"--> Processing volume {volume}...")
        assert_array_almost_equal(
            np.asarray(img1.dataobj[..., volume]),
            np.asarray(img2.dataobj[..., volume]),
            decimal=decimal,
        )


assert_nifti_equal = partial(
    _assert_nifti_relation,
    assertion_func_data=_assert_dataobj_equal,
    assertion_func_affine=_assert_affine_equal,
)


assert_nifti_almost_equal = partial(
    _assert_nifti_relation,
    assertion_func_data=_assert_dataobj_almost_equal,
    assertion_func_affine=_assert_affine_almost_equal,
)


assert_large_nifti_almost_equal = partial(
    _assert_nifti_relation,
    assertion_func_data=_assert_large_image_dataobj_almost_equal,
    assertion_func_affine=_assert_affine_almost_equal,
)


def build_dwi_dataset(
    folder: Path, nb_dwi_volumes: int, nb_b_values: int, nb_b_vectors: int
) -> DWIDataset:
    dwi_data = 4.0 * np.ones((5, 5, 5, nb_dwi_volumes))
    dwi_img = nib.Nifti1Image(dwi_data, affine=np.eye(4))
    nib.save(dwi_img, folder / "foo.nii.gz")
    np.savetxt(folder / "foo.bval", [1000] * nb_b_values)
    b_vectors_data = np.random.random((3, nb_b_vectors))
    np.savetxt(folder / "foo.bvec", b_vectors_data)

    return DWIDataset(
        dwi=folder / "foo.nii.gz",
        b_values=folder / "foo.bval",
        b_vectors=folder / "foo.bvec",
    )



"""Module for converting fMRI of ADNI."""

from pathlib import Path
from typing import Any, Iterable, Optional

import pandas as pd

__all__ = ["convert_fmri"]


def convert_fmri(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
    convert_multiband: bool = True,
):
    """Convert fMR images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    destination_dir : Path
        The path to the destination BIDS directory.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.

    convert_multiband : bool, optional
        If True, the converter will also convert multiband fmri scans.
        Otherwise, they will be ignored.
        Default=True.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        ADNIModalityConverter,
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        f"Calculating paths of {ADNIModalityConverter.FMRI} images. Output will be stored in {conversion_dir}.",
        lvl="info",
    )
    images = _compute_fmri_path(
        source_dir, csv_dir, subjects, conversion_dir, convert_multiband
    )
    cprint(
        f"Paths of {ADNIModalityConverter.FMRI} images found. Exporting images into BIDS ...",
        lvl="info",
    )
    paths_to_bids(
        images,
        destination_dir,
        ADNIModalityConverter.FMRI,
        mod_to_update=mod_to_update,
        n_procs=n_procs,
    )
    cprint(msg=f"{ADNIModalityConverter.FMRI} conversion done.", lvl="debug")


def _compute_fmri_path(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
    convert_multiband: bool,
) -> pd.DataFrame:
    """Compute the paths to fMR images.

    Parameters
    ----------
    source_dir : Path
        The path to the ADNI directory.

    csv_dir : Path
        The path to the clinical data directory.

    subjects : list of str
        The list of subjects.

    conversion_dir : Path
        The path to the TSV files including the paths to original images.

    convert_multiband : bool, optional
        If True, the converter will also convert multiband fmri scans.
        Otherwise, they will be ignored.
        Default=True.

    Returns
    -------
    images : pd.DataFrame
        Pandas Dataframe containing the path for each fmri.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv

    from ._image_path_utils import find_image_path
    from ._visits_utils import visits_to_timepoints

    fmri_dfs_list = []
    fmri_df = _initialize_fmri_df()
    adni_merge = load_clinical_csv(csv_dir, "ADNIMERGE")
    mayo_mri_qc = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15")
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "fMRI"]
    mayo_mri_qc.columns = [x.upper() for x in mayo_mri_qc.columns]
    mayo_mri_qc3 = load_clinical_csv(csv_dir, "MAYOADIRL_MRI_QUALITY_ADNI3")
    mayo_mri_qc3 = mayo_mri_qc3[mayo_mri_qc3.SERIES_TYPE == "EPB"]

    # Concatenating visits in both QC files
    mayo_mri_qc = pd.concat(
        [mayo_mri_qc, mayo_mri_qc3], axis=0, ignore_index=True, sort=False
    )

    mri_list = load_clinical_csv(csv_dir, "MRILIST")

    # Selecting fMRI images
    mri_list = mri_list[mri_list.SEQUENCE.str.contains("MRI")]
    unwanted_sequences = [] if convert_multiband else ["MB"]
    mri_list = mri_list[
        mri_list.SEQUENCE.map(
            lambda x: not any(subs in x for subs in unwanted_sequences)
        )
    ]
    # We will convert the images for each subject in the subject list
    for subject in subjects:
        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subject]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")
        mri_list_subj = mri_list[mri_list.SUBJECT == subject]
        mri_list_subj = mri_list_subj.sort_values("SCANDATE")
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subject[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(
            subject, mri_list_subj, adnimerge_subj, modality="fMRI"
        )

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]
            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            selected_image = _select_fmri_image(
                subject, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj
            )
            if selected_image is not None:
                fmri_dfs_list.append(pd.DataFrame(selected_image, index=["i"]))
    if fmri_dfs_list:
        fmri_df = pd.concat(fmri_dfs_list, ignore_index=True)

    # Removing known exceptions from images to convert
    if not fmri_df.empty:
        error_ind = fmri_df.index[
            fmri_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in _get_known_conversion_errors()),
                axis=1,
            )
        ]
        fmri_df.drop(error_ind, inplace=True)
    # Checking for images paths in filesystem
    images = find_image_path(fmri_df, source_dir, modality="fMRI")
    images.to_csv(conversion_dir / "fmri_paths.tsv", sep="\t", index=False)

    return images


def _initialize_fmri_df() -> pd.DataFrame:
    return pd.DataFrame(
        columns=[
            "Subject_ID",
            "VISCODE",
            "Visit",
            "Sequence",
            "Scan_Date",
            "Study_ID",
            "Field_Strength",
            "Series_ID",
            "Image_ID",
        ]
    )


def _get_known_conversion_errors() -> Iterable[tuple[str, str]]:
    return [
        ("006_S_4485", "m84"),
        ("123_S_4127", "m96"),
        # Eq_1
        ("094_S_4503", "m24"),
        ("009_S_4388", "m72"),
        ("036_S_6088", "bl"),
        ("036_S_6134", "bl"),
        ("016_S_6802", "bl"),
        ("016_S_6816", "bl"),
        ("126_S_4891", "m84"),
        # Multiple images
        ("029_S_2395", "m72"),
    ]


def _select_fmri_image(
    subject_id: str,
    timepoint: str,
    visit_str: str,
    visit_mri_list: pd.DataFrame,
    mri_qc_subj: pd.DataFrame,
) -> Optional[dict[str, Any]]:
    """Select the fMRI image for the given subject and visit based on QC.

    One image among those in the input list is chosen according to QC
    and then corresponding metadata is extracted to a dictionary.

    Parameters
    ----------
    subject_id : str
        The subject identifier.

    timepoint : str
        The visit code.

    visit_str : str
        The visit name.

    visit_mri_list : pd.DataFrame
        List of images metadata.

    mri_qc_subj : pd.DataFrame
        A Dataframe containing list of QC of scans for the subject.

    Returns
    -------
    dict or None :
        A dictionary which contains image metadata.
        If None, no image was found.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    from ._qc_utils import select_image_qc

    mri_qc_subj.columns = [x.lower() for x in mri_qc_subj.columns]
    selected_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if not selected_image:
        return None
    selected_scan = visit_mri_list[visit_mri_list.IMAGEUID == selected_image].iloc[0]

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": replace_sequence_chars(selected_scan.SEQUENCE),
        "Scan_Date": selected_scan["SCANDATE"],
        "Study_ID": str(int(selected_scan.STUDYID)),
        "Series_ID": str(int(selected_scan.SERIESID)),
        "Image_ID": str(int(selected_scan.IMAGEUID)),
        "Field_Strength": selected_scan.MAGSTRENGTH,
    }



from pathlib import Path

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["validate_input_path"]


def validate_input_path(input_path: UserProvidedPath, check_exist: bool = True) -> Path:
    """Take a user provided path as input and returns a resolved, existing if check_exist is True, Path."""
    input_path = Path(input_path).resolve()
    if check_exist and not input_path.exists():
        raise FileNotFoundError(f"The provided input path {input_path} does not exist.")
    return input_path



import typing as ty

import nipype.interfaces.spm as spm
import pydra
from nipype.algorithms.misc import Gunzip
from pydra import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_utils import (
    ApplySegmentationDeformation,
)
from clinica.pydra.engine import clinica_io
from clinica.pydra.t1_volume.utils import initialize_tissues_spm_segment


@clinica_io
def t1volume_tissue_segmentation(
    name: str = "t1-volume-tissue-segmentation", parameters: dict = {}
) -> Workflow:
    """Core workflow for tissue segmentation, bias correction and spatial normalization

    Parameters
    ----------
    name : str
        name of pipeline. Default="t1-volume-tissue-segmentation".

    parameters : dict
        dictionary of pipeline parameters

    Returns
    -------
    Workflow
        pydra workflow for core functionalities
    """

    from clinica.pydra.t1_volume.tasks import (
        task_volume_location_in_world_coordinate_system,
    )

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[("T1w", str, {"mandatory": True})],
        bases=(pydra.specs.BaseSpec,),
    )

    workflow = Workflow(
        name,
        input_spec=input_spec,
    )

    workflow.split("T1w")

    workflow.add(
        task_volume_location_in_world_coordinate_system(
            name="check_location_world",
            nifti_input=workflow.lzin.T1w,
            skip_question=parameters["skip_question"],
        )
    )

    workflow.add(
        Nipype1Task(name="unzipT1w", interface=Gunzip(), in_file=workflow.lzin.T1w)
    )

    spm_segment = Nipype1Task(
        name="SpmSegmentation",
        interface=spm.NewSegment(),
    )

    spm_segment.inputs.write_deformation_fields = [True, True]

    tissue_tuples = initialize_tissues_spm_segment(parameters)
    spm_segment.inputs.tissues = tissue_tuples

    spm_segment.inputs.channel_files = workflow.unzipT1w.lzout.out_file

    workflow.add(spm_segment)

    t1_to_mni = Nipype1Task(
        name="T1wToMNI",
        interface=ApplySegmentationDeformation(),
    )

    t1_to_mni.inputs.in_files = workflow.unzipT1w.lzout.out_file
    t1_to_mni.inputs.deformation_field = (
        workflow.SpmSegmentation.lzout.forward_deformation_field
    )

    workflow.add(t1_to_mni)

    workflow.set_output([("test_out", workflow.T1wToMNI.lzout.out_files)])

    return workflow



import functools
from os import PathLike

from pydra import Workflow
from pydra.engine.core import TaskBase

import clinica.pydra.engine_utils as pu
from clinica.pydra.interfaces import bids_reader, bids_writer, caps_reader


def clinica_io(func):
    """Decorator to add BIDS reader/writer to any Pydra workflow."""

    @functools.wraps(func)
    def run_wrapper(
        name: str, parameters: dict, input_dir: PathLike, output_dir: PathLike
    ) -> Workflow:
        core_workflow = func(parameters=parameters)

        pipeline = Workflow(
            name=name,
            input_spec=["bids_dir", "caps_dir"],
            bids_dir=input_dir,
            caps_dir=output_dir,
        )

        build_input_workflow(pipeline, core_workflow)

        # TODO: define condition on split if multiple fields
        pipeline.add(core_workflow)

        build_output_workflow(pipeline, core_workflow, output_dir)

        return pipeline

    return run_wrapper


def add_input_reading_task(
    pipeline: Workflow,
    core_workflow: Workflow,
    query_type: str,
    reader: TaskBase,
) -> Workflow:
    """Configure and add the reading tasks of input workflow.

    Parameters
    ----------
    pipeline : Workflow
        The main Workflow to which the readers should be added.

    core_workflow : Workflow
        The core workflow. It defines, through its inputs,
        what the reader workflow should read and these two
        workflows should connect.

    query_type : {"bids", "caps_file", "caps_group"}
        The type of query that should be run.

    reader : TaskBase
        Task responsible for reading data.

    Returns
    -------
    pipeline : Workflow
        The main Workflow with readers added to it.
    """
    from clinica.pydra.query import query_factory

    query = query_factory(pu.list_workflow_inputs(core_workflow), query_type=query_type)
    if len(query) == 0:
        return pipeline
    input_dir = "bids_dir" if "bids" in reader.__name__ else "caps_dir"
    input_workflow = Workflow(
        name=f"input_workflow_{query_type}_reader",
        input_spec=["input_dir"],
    )
    try:
        input_workflow.inputs.input_dir = getattr(pipeline.lzin, input_dir)
    except AttributeError:
        raise AttributeError(
            f"Workflow has no {input_dir} input. Please verify your input specifications."
        )
    input_workflow = add_input_task(
        input_workflow,
        reader(query=query, input_dir=input_workflow.lzin.input_dir),
    )
    pipeline.add(input_workflow)

    # connect input workflow to core workflow
    for field, _ in query.query.items():
        read_data = getattr(input_workflow.lzout, field)
        setattr(core_workflow.inputs, field, read_data)

    return pipeline


add_input_reading_task_bids = functools.partial(
    add_input_reading_task,
    query_type="bids",
    reader=bids_reader,
)


add_input_reading_task_caps_file = functools.partial(
    add_input_reading_task,
    query_type="caps_file",
    reader=caps_reader,
)


add_input_reading_task_caps_group = functools.partial(
    add_input_reading_task,
    query_type="caps_group",
    reader=caps_reader,
)


def build_input_workflow(pipeline: Workflow, core_workflow: Workflow) -> Workflow:
    """Setup for an input workflow.

    For now, the input workflow is responsible for:

        - reading BIDS data
        - reading CAPS data with clinica_file_reader
        - reading CAPS data with clinica_group_reader

    Parameters
    ----------
    pipeline :  Workflow
         The high level workflow containing (input -> core -> output).

    core_workflow : Workflow
         The functional workflow.

    Returns
    -------
    pipeline : Workflow
        The pipeline with the input workflow.
    """
    pipeline = add_input_reading_task_bids(pipeline, core_workflow)
    pipeline = add_input_reading_task_caps_file(pipeline, core_workflow)
    pipeline = add_input_reading_task_caps_group(pipeline, core_workflow)
    return pipeline


def add_input_task(input_workflow: Workflow, task: TaskBase) -> Workflow:
    """Add a task to the input workflow and define the workflow outputs.

    Parameters
    ----------
    input_workflow : Workflow
        The high level workflow containing (input -> core -> output).

    task : TaskBase
        The task to be added to the input_workflow.

    Returns
    -------
    input_workflow : Workflow
         The input workflow with the task added to it.
    """
    input_workflow.add(task)
    input_workflow.set_output(
        [
            (
                field,
                getattr(getattr(input_workflow, f"{task.name}").lzout, field),
            )
            for field in task.output_names
        ]
    )
    return input_workflow


def build_output_workflow(
    pipeline: Workflow, core_workflow: Workflow, output_dir: PathLike
) -> Workflow:
    """Example of an output workflow.

    Parameters
    ----------
    pipeline : Workflow
        the resulting workflow consisting of (input/core/output)
    core_workflow : Workflow
        contains the core interfaces
    output_dir : PathLike
        Path of the directory to be written to

    Returns
    -------
    Workflow
        The pipeline with the output workflow.
    """

    output_attrs = []

    for i, field in enumerate(pu.list_out_fields(core_workflow)):
        pipeline.add(bids_writer(name="bids_writer_task_" + str(field)))

        writer_task = getattr(pipeline, "bids_writer_task_" + str(field))
        writer_task_inputs = getattr(writer_task, "inputs")

        output_data = getattr(core_workflow.lzout, field)

        setattr(writer_task_inputs, "output_file", output_data)
        setattr(writer_task_inputs, "output_dir", output_dir)

        output_attrs.append((field, output_data))

    pipeline.set_output(output_attrs)

    return pipeline



from typing import Optional

import pandas as pd

from clinica.iotools.converters.adni_to_bids.adni_utils import ADNIStudy
from clinica.utils.stream import cprint

__all__ = ["visits_to_timepoints"]


def visits_to_timepoints(
    subject: str,
    mri_list_subj: pd.DataFrame,
    adnimerge_subj: pd.DataFrame,
    modality: str,
    visit_field: str = "VISIT",
    scandate_field: str = "SCANDATE",
) -> dict:
    """
    A correspondence is established between ADNIMERGE and MRILIST visits, since they have different notations.

    Args:
        subject: Subject identifier
        mri_list_subj: Dataframe containing list of MRI scans for the subject
        adnimerge_subj: Dataframe containing visits data for the subject
        modality: Imaging modality
        visit_field: field name corresponding to the visit
        scandate_field: field name corresponding to the scan date

    Returns:
        [Returns]
    """
    if modality == "T1":
        mri_list_subj = mri_list_subj[mri_list_subj[visit_field] != "ADNI Baseline"]

    visits = dict()
    unique_visits = list(mri_list_subj[visit_field].unique())
    pending_timepoints = []

    # We try to obtain the corresponding image Visit for a given VISCODE
    for adni_row in adnimerge_subj.iterrows():
        visit = adni_row[1]
        study = ADNIStudy(visit.ORIGPROT)
        preferred_visit_name = _get_preferred_visit_name(study, visit.VISCODE)

        if preferred_visit_name in unique_visits:
            key_preferred_visit = (visit.VISCODE, visit.COLPROT, visit.ORIGPROT)
            if key_preferred_visit not in visits:
                visits[key_preferred_visit] = preferred_visit_name
            elif visits[key_preferred_visit] != preferred_visit_name:
                cprint(
                    f"[{modality}] Subject {subject} has multiple visits for one timepoint.",
                    lvl="info",
                )

            unique_visits.remove(preferred_visit_name)
            continue

        pending_timepoints.append(visit)

    # Then for images.Visit non matching the expected labels we find the closest date in visits list
    for visit in unique_visits:
        image = (mri_list_subj[mri_list_subj[visit_field] == visit]).iloc[0]

        closest_visit = _get_closest_visit(
            image[scandate_field],
            image[visit_field],
            pending_timepoints,
            subject,
        )
        if closest_visit is None:
            cprint(f"No closest visit found for image {image}", lvl="debug")
            continue
        key_min_visit = (
            closest_visit.VISCODE,
            closest_visit.COLPROT,
            closest_visit.ORIGPROT,
        )

        if key_min_visit not in visits.keys():
            visits[key_min_visit] = image[visit_field]
        elif visits[key_min_visit] != image[visit_field]:
            cprint(
                f"[{modality}] Subject {subject} has multiple visits for one timepoint.",
                lvl="debug",
            )

    return visits


def _get_preferred_visit_name(study: ADNIStudy, visit_code: str) -> str:
    """Return the expected visit name for a given visit depending
    on the study (ADNI1, ADNI2, ADNI3, ADNIGO).

    Parameters
    ----------
    study : ADNIStudy
        The study for this visit.
    visit_code : str
        The visit code to convert to a new form.

    Returns
    -------
    str :
        The expected visit name.
    """
    if study == ADNIStudy.ADNI3:
        return _get_preferred_visit_name_adni3(visit_code)
    if study == ADNIStudy.ADNI2:
        return _get_preferred_visit_name_adni2(visit_code)
    if study == ADNIStudy.ADNI1:
        return _get_preferred_visit_name_adni1(visit_code)
    if study == ADNIStudy.ADNIGO:
        return _get_preferred_visit_name_adnigo(visit_code)


def _get_preferred_visit_name_adni3(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI Screening"
    return f"ADNI3 Year {_parse_year_from_visit_code(visit_code)} Visit"


def _parse_year_from_visit_code(visit_code: str) -> float:
    """Return the year corresponding to the visit code.
    Assumes a visit code of the form 'mXXX'.
    """
    return _parse_month_from_visit_code(visit_code) / 12


def _parse_month_from_visit_code(visit_code: str) -> int:
    """Return the month corresponding to the visit code.
    Assumes a visit code of the form 'mXXX'.
    """
    try:
        return int(visit_code[1:])
    except Exception:
        raise ValueError(
            f"Cannot extract month from visit code {visit_code}."
            "Expected a code of the form 'mXXX' where XXX can be casted to an integer."
        )


def _get_preferred_visit_name_adni2(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI2 Screening MRI-New Pt"
    if visit_code == "m03":
        return "ADNI2 Month 3 MRI-New Pt"
    if visit_code == "m06":
        return "ADNI2 Month 6-New Pt"
    return f"ADNI2 Year {_parse_year_from_visit_code(visit_code)} Visit"


def _get_preferred_visit_name_adni1(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNI Screening"
    if visit_code == "m03":
        return "ADNIGO Month 3 MRI"  # does not make any sense...
    month = _parse_month_from_visit_code(visit_code)
    return f"ADNI{'1/' if month < 54 else ''}GO Month {month}"


def _get_preferred_visit_name_adnigo(visit_code: str) -> str:
    if visit_code == "bl":
        return "ADNIGO Screening MRI"
    return _get_preferred_visit_name_adni1(visit_code)


def _get_closest_visit(
    image_acquisition_date: str,
    image_visit: str,
    visits: list[pd.Series],
    subject: str,
) -> Optional[pd.Series]:
    """Choose the visit with the closest date to a given image acquisition date.

    Parameters
    ----------
    image_acquisition_date : str
        The date in string format when the image was acquired.

    image_visit : str
        The string identifier of the visit.

    visits : list of pd.Series
        List of visit entries among which we have to find the closest one.

    subject : str
        Subject identifier.

    Returns
    -------
    None or pd.Series
        If the list of visits provided is empty, this function returns None.
        Otherwise, it returns a pd.Series corresponding to the closest visit found.
    """
    if len(visits) == 0:
        cprint(
            "No corresponding timepoint in ADNIMERGE for "
            f"subject {subject} in visit {image_visit}",
            lvl="info",
        )
        return None

    if len(visits) == 1:
        return visits[0]

    (
        closest_visit,
        second_closest_visit,
        smallest_time_difference,
        second_smallest_time_difference,
    ) = _get_closest_and_second_closest_visits(visits, image_acquisition_date)

    return (
        _get_closest_visit_for_large_time_difference(
            subject,
            image_visit,
            image_acquisition_date,
            closest_visit,
            second_closest_visit,
            smallest_time_difference,
            second_smallest_time_difference,
        )
        if smallest_time_difference > 90
        else closest_visit
    )


def _get_closest_visit_for_large_time_difference(
    subject: str,
    image_visit: str,
    image_acquisition_date: str,
    closest_visit: pd.Series,
    second_closest_visit: pd.Series,
    smallest_time_difference: int,
    second_smallest_time_difference: int,
) -> pd.Series:
    """Special case when the closest visit found is more than 90 days
    apart from the image acquisition date.

    In this situation, if the image acquisition date is in between the
    closest visit and the second-closest visit, and the second-closest
    visit is close enough from the image acquisition date, then select
    the second-closest visit instead of the closest one.
    """
    cprint(
        _get_smallest_time_difference_too_large_message(
            subject,
            image_visit,
            image_acquisition_date,
            closest_visit,
            second_closest_visit,
            smallest_time_difference,
            second_smallest_time_difference,
        ),
        lvl="debug",
    )
    if _second_closest_visit_is_better(
        closest_visit.EXAMDATE,
        image_acquisition_date,
        second_closest_visit.EXAMDATE,
    ):
        closest_visit = second_closest_visit
    cprint(msg=f"We prefer {closest_visit.VISCODE}", lvl="debug")

    return closest_visit


def _second_closest_visit_is_better(
    closest_visit_date: str,
    image_acquisition_date: str,
    second_closest_visit_date: str,
) -> bool:
    """If image is too close to the date between two visits we prefer the earlier visit."""
    from datetime import datetime

    date_format = "%Y-%m-%d"
    if (
        datetime.strptime(closest_visit_date, date_format)
        > datetime.strptime(image_acquisition_date, date_format)
        > datetime.strptime(second_closest_visit_date, date_format)
    ):
        diff = _compute_number_of_days_between(
            closest_visit_date, second_closest_visit_date
        )
        smallest_time_difference = _compute_number_of_days_between(
            image_acquisition_date, closest_visit_date
        )
        if abs((diff / 2.0) - smallest_time_difference) < 30:
            return True
    return False


def _get_closest_and_second_closest_visits(
    visits: list[pd.Series],
    reference_date: str,
) -> tuple[pd.Series, pd.Series, int, int]:
    """Return the closest and second-closest visit from the reference date.

    Parameters
    ----------
    visits : list of pd.Series
        The list of visits to analyze. Should have a length greater than two.

    reference_date : str
        The reference date used to compute the time differences.

    Returns
    -------
    closest_visit : pd.Series
        The visit whose EXAMDATE is the closest to the reference date.

    second_closest_visit : pd.Series
        The second closest visit to the reference date.

    min_time_difference : int
        The number of days between the reference date and the closest visit.

    second_min_time_difference : int
        The number of days between the reference date and the second closest visit.
    """
    import numpy as np

    if len(visits) < 2:
        raise ValueError(f"Expected at least two visits, received {len(visits)}.")
    time_differences_between_reference_and_visits_in_days = [
        _compute_number_of_days_between(reference_date, visit_date)
        for visit_date in [visit.EXAMDATE for visit in visits]
    ]
    idx = np.argsort(time_differences_between_reference_and_visits_in_days)
    min_time_difference = time_differences_between_reference_and_visits_in_days[idx[0]]
    second_min_time_difference = time_differences_between_reference_and_visits_in_days[
        idx[1]
    ]

    return (
        visits[idx[0]],
        visits[idx[1]],
        min_time_difference,
        second_min_time_difference,
    )


def _get_smallest_time_difference_too_large_message(
    subject: str,
    image_visit: str,
    image_acquisition_date: str,
    closest_visit: pd.Series,
    second_closest_visit: pd.Series,
    smallest_time_difference: int,
    second_smallest_time_difference: int,
) -> str:
    """Build the debug message when the closest found visit is further than 90 days in time."""
    msg = (
        f"More than 90 days for corresponding timepoint in ADNIMERGE for "
        f"subject {subject} in visit {image_visit} on {image_acquisition_date}.\n"
    )
    idx = 1
    for visit, difference in zip(
        [closest_visit, second_closest_visit],
        [smallest_time_difference, second_smallest_time_difference],
    ):
        msg += (
            f"Timepoint {idx}: {visit.VISCODE} - {visit.ORIGPROT} "
            f"on {visit.EXAMDATE} (Distance: {difference} days)\n"
        )
        idx += 1
    return msg


def _compute_number_of_days_between(date_1: str, date_2: str) -> int:
    """Calculate the days between two dates."""
    from datetime import datetime

    date_1 = datetime.strptime(date_1, "%Y-%m-%d")
    date_2 = datetime.strptime(date_2, "%Y-%m-%d")
    return abs((date_2 - date_1).days)



from pathlib import Path
from typing import Iterator, Optional

import pandas as pd

from clinica.iotools.bids_utils import StudyName, bids_id_factory
from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def _get_protocol_to_bids_df() -> pd.DataFrame:
    return pd.DataFrame.from_dict(
        {
            "3DFLAIR": {"datatype": "anat", "modality": "FLAIR"},
            "ADNI_1X-MPRAGE": {"datatype": "anat", "modality": "T1w"},
            "ADNI_2X-MPRAGE": {"datatype": "anat", "modality": "T1w"},
            "FDG": {
                "datatype": "pet",
                "modality": "pet",
                "trc_label": "18FFDG",
            },
            "FTP_75-105": {
                "datatype": "pet",
                "modality": "pet",
                "trc_label": "18FFTP",
            },
            "FTP_80-100": {
                "datatype": "pet",
                "modality": "pet",
                "trc_label": "18FFTP",
            },
            "PIB_40-60": {
                "datatype": "pet",
                "modality": "pet",
                "trc_label": "11CPIB",
            },
        },
        orient="index",
    )


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    *args,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    from clinica.iotools.bids_utils import StudyName
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.HABS)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.HABS)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )
    clinical_data = {
        k: _read_clinical_data(path_to_dataset / p, c)
        for k, p, c in _find_clinical_data(path_to_dataset)
    }
    imaging_data = pd.concat(
        [_parse_imaging_data(x) for x in _find_imaging_data(path_to_dataset)]
    )
    _write_bids(
        sourcedata=path_to_dataset,
        rawdata=bids_dir,
        imaging_data=imaging_data,
        clinical_data=clinical_data,
    )


def _source_session_id_to_bids(dataframe: pd.DataFrame) -> pd.Series:
    import re

    # HABS session format as `{years}.{months}`
    years_dot_months_pattern = r".*_(\d+).(\d+)"

    def replace_years_dot_months_session(match: re.Match) -> str:
        years = int(match.group(1)) - 1
        months = int(match.group(2))
        return f"ses-M{12 * years + months:03d}"

    # HABS undocumented format as `{months}m`
    months_only_pattern = r".*_(\d+)m.*"

    def replace_months_only_pattern(match: re.Match) -> str:
        months = int(match.group(1))
        return f"ses-M{months:03d}"

    return dataframe.source_session_id.str.replace(
        years_dot_months_pattern, replace_years_dot_months_session, regex=True
    ).str.replace(months_only_pattern, replace_months_only_pattern, regex=True)


def _find_clinical_data(sourcedata: Path) -> Iterator[tuple[str, Path, dict[str, str]]]:
    from csv import DictReader

    for f in sourcedata.rglob("*HABS_DataRelease*.csv"):
        header = DictReader(open(f, encoding="utf-8-sig")).fieldnames
        yield (
            f.name.split("_")[0],
            f.relative_to(sourcedata),
            {
                header[0]: "source_participant_id",
                header[1]: "source_session_id",
                header[2]: "date",
            },
        )


def _read_clinical_data(path: Path, rename_columns: dict[str, str]) -> pd.DataFrame:
    return (
        pd.read_csv(path)
        .rename(columns=rename_columns)
        .assign(date=lambda df: pd.to_datetime(df.date))
        .assign(
            participant_id=lambda df: df.source_participant_id.apply(
                lambda x: bids_id_factory(StudyName.HABS).from_original_study_id(x)
            )
        )
        .drop(columns="source_participant_id")
        .assign(session_id=_source_session_id_to_bids)
        .drop(columns="source_session_id")
        .convert_dtypes()
        .set_index(["participant_id", "session_id", "date"], verify_integrity=True)
        .sort_index()
    )


def _find_imaging_data(sourcedata: Path) -> Iterator[list[tuple[str, str]]]:
    from zipfile import ZipFile

    for z in sourcedata.rglob("*HABS_DataRelease*.zip"):
        yield [
            (str(z.relative_to(sourcedata)), f)
            for f in ZipFile(z).namelist()
            if f.endswith(".nii.gz")
        ]


def _parse_imaging_data(paths: list[tuple[str, str]]) -> Optional[pd.DataFrame]:
    df = pd.DataFrame.from_records(paths, columns=["source_zipfile", "source_filename"])
    # Parse imaging metadata from file paths.
    pattern = (
        r"(?P<source_participant_id>P_\w{6})"
        r"_(?P<date>[\d-]+)"
        r"_(?P<source_session_id>HAB_\d{1}.\d{1,2})"
        r"_(?P<protocol>[\w-]+)"
    )
    df = pd.concat(
        [df, df["source_filename"].str.extract(pattern)],
        axis="columns",
    ).assign(date=lambda x: pd.to_datetime(x.date))

    # Map protocol to BIDS entities.
    df = (
        df.join(_get_protocol_to_bids_df(), on="protocol", how="left")
        .drop(columns="protocol")
        .dropna(subset=["datatype", "modality"])
    )
    if df.empty:
        return None
    # Compute BIDS participant ID, session ID and filename.
    df = (
        df.assign(
            participant_id=lambda df: df.source_participant_id.apply(
                lambda x: bids_id_factory(StudyName.HABS).from_original_study_id(x)
            )
        )
        .drop(columns="source_participant_id")
        .assign(session_id=_source_session_id_to_bids)
        .drop(columns="source_session_id")
        .set_index(
            ["participant_id", "session_id", "datatype", "modality", "trc_label"],
            verify_integrity=True,
        )
        .sort_index()
    )

    return df


def _install_nifti(zipfile: str, filename: str, bids_path: str) -> None:
    """Install a NIfTI file from a source archive to the target BIDS path."""
    import fsspec

    fo = fsspec.open(zipfile)
    fs = fsspec.filesystem("zip", fo=fo)
    with fsspec.open(bids_path, mode="wb") as f:
        f.write(fs.cat(filename))


def _write_bids(
    sourcedata: Path,
    rawdata: Path,
    imaging_data: pd.DataFrame,
    clinical_data: dict[str, pd.DataFrame],
):
    import fsspec
    from pandas import notna

    from clinica.iotools.bids_dataset_description import BIDSDatasetDescription
    from clinica.iotools.bids_utils import (
        StudyName,
        write_modality_agnostic_files,
        write_to_tsv,
    )

    participants = (
        clinical_data["Demographics"]
        .drop(columns="NP_Age")
        .rename(
            columns={
                "BiologicalSex": "sex",
                "YrsOfEd": "years_of_education",
            }
        )
        .xs("ses-M000", level="session_id")
        .reset_index(level="date", drop=True)
    ).rename(columns=str.lower)

    with fsspec.open(str(rawdata / "dataset_description.json"), mode="wt") as f:
        BIDSDatasetDescription(name=StudyName.HABS).write(to=f)

    participants_file = rawdata / "participants.tsv"
    with fsspec.open(str(participants_file), mode="wb") as f:
        write_to_tsv(participants, f)

    sessions = (
        clinical_data["Demographics"][["NP_Age"]]
        .reset_index(level="date", drop=True)
        .rename(columns={"NP_Age": "age"})
        .join(
            clinical_data["ClinicalMeasures"]
            .rename(columns={"HABS_DX": "diagnostic"})
            .reset_index(level="date", drop=True)
        )
    ).rename(columns=str.lower)

    for participant_id, dataframe in sessions.groupby("participant_id"):
        sessions_file = rawdata / participant_id / f"{participant_id}_sessions.tsv"
        dataframe = dataframe.droplevel("participant_id")
        with fsspec.open(sessions_file, mode="wb") as f:
            write_to_tsv(dataframe, f)

    for grouped_by, dataframe in imaging_data.groupby(["participant_id", "session_id"]):
        participant_id, session_id = grouped_by
        bids_basedir = rawdata / participant_id / session_id
        bids_prefix = f"{participant_id}_{session_id}"
        dataframe = (
            dataframe.droplevel(level=["participant_id", "session_id"])
            .reset_index()
            .assign(
                filename=lambda df: df.apply(
                    lambda row: (
                        f"{row.datatype}/{bids_prefix}"
                        f"{'_trc-' + row.trc_label if notna(row.trc_label) else ''}"
                        f"_{row.modality}.nii.gz"
                    ),
                    axis="columns",
                )
            )
            .drop(columns=["datatype", "modality", "trc_label"])
            .set_index("filename", verify_integrity=True)
        )

        for filename, row in dataframe.iterrows():
            _install_nifti(
                zipfile=str(sourcedata / row.source_zipfile),
                filename=row.source_filename,
                bids_path=str(bids_basedir / filename),
            )

        dataframe = dataframe.drop(columns=["source_zipfile", "source_filename"])

        with fsspec.open(bids_basedir / f"{bids_prefix}_scans.tsv", mode="wb") as f:
            write_to_tsv(dataframe, f)

    readme_data = {
        "link": "https://habs.mgh.harvard.edu",
        "desc": (
            "The overall goal of the Harvard Aging Brain Study (HABS) is to elucidate the earliest changes in "
            "molecular, functional and structural imaging markers that signal the transition from normal cognition to "
            "progressive cognitive decline along the trajectory of preclinical Alzheimer’s Disease."
        ),
    }
    write_modality_agnostic_files(
        study_name=StudyName.HABS,
        readme_data=readme_data,
        bids_dir=rawdata,
    )



import nibabel as nib
import numpy as np


def load_data(image_list, mask=True):
    """

    Args:
        image_list:
        mask:

    Returns:

    """
    data = None
    shape = None
    data_mask = None
    first = True

    for i in range(len(image_list)):
        subj = nib.load(image_list[i])
        subj_data = np.nan_to_num(subj.get_fdata(dtype="float32").flatten())

        # Memory allocation for ndarray containing all data to avoid copying the array for each new subject
        if first:
            data = np.ndarray(
                shape=(len(image_list), subj_data.shape[0]), dtype=float, order="C"
            )
            shape = subj.get_fdata(dtype="float32").shape
            first = False

        data[i, :] = subj_data

    if mask:
        data_mask = (data != 0).sum(axis=0) != 0
        data = data[:, data_mask]

    return data, shape, data_mask


def revert_mask(weights, mask, shape):
    """

    Args:
        weights:
        mask:
        shape:

    Returns:

    """

    z = np.zeros(np.prod(shape))
    z[mask] = weights

    new_weights = np.reshape(z, shape)

    return new_weights


def features_weights(image_list, dual_coefficients, sv_indices, scaler=None, mask=None):
    if len(sv_indices) != len(dual_coefficients):
        print("Length dual coefficients: " + str(len(dual_coefficients)))
        print("Length indices: " + str(len(sv_indices)))
        raise ValueError(
            "The number of support vectors indices and the number of coefficients must be the same."
        )

    if len(image_list) == 0:
        raise ValueError("The number of images must be greater than 0.")

    sv_images = [image_list[i] for i in sv_indices]

    shape = nib.load(sv_images[0]).get_fdata(dtype="float32").shape
    weights = np.zeros(shape)

    for i in range(len(sv_images)):
        subj = nib.load(sv_images[i])
        subj_data = np.nan_to_num(subj.get_fdata(dtype="float32"))

        if scaler is not None and mask is not None:
            subj_data = subj_data.flatten()[mask]
            subj_data = scaler.transform(subj_data)
            subj_data = revert_mask(subj_data, mask, shape)
        weights += dual_coefficients[i] * subj_data

    return weights


def weights_to_nifti(weights, image, output_filename):
    """

    Args:
        weights:
        image:
        output_filename:

    Returns:

    """

    # Normalize with 2-norm
    # features = 2 * weights / np.power(norm(weights.flatten(), 2), 2)

    # Normalize inf-norm
    features = weights / abs(weights).max()

    img = nib.load(image)
    canonical_img = nib.as_closest_canonical(img)
    hd = canonical_img.header
    qform = np.zeros((4, 4))

    for i in range(1, 4):
        qform[i - 1, i - 1] = hd["pixdim"][i]
        qform[i - 1, 3] = -1.0 * hd["pixdim"][i] * hd["dim"][i] / 2.0

    output_image = nib.Nifti1Image(features, qform)
    nib.save(output_image, output_filename)



from typing import List, Optional

import pandas as pd

from ._base import GLM
from ._contrast import CorrelationContrast


class CorrelationGLM(GLM):
    """Class implementing the correlation type GLM model.

    Attributes
    ----------
    See documentation for `GLM` class.

    group_label : str, optional
        The label to use for group GLM models. Default=None.
    """

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        group_label: Optional[str],
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = 0.001,
        threshold_corrected_pvalue: Optional[float] = 0.05,
        cluster_threshold: Optional[float] = 0.001,
    ):
        super().__init__(
            design,
            df,
            feature_label,
            contrast,
            fwhm,
            threshold_uncorrected_pvalue,
            threshold_corrected_pvalue,
            cluster_threshold,
        )
        self.with_interaction: bool = False
        self.group_label = group_label

    def _build_contrasts(
        self, contrast: str, subjects_df: pd.DataFrame
    ) -> List[CorrelationContrast]:
        return [CorrelationContrast.from_string(contrast, subjects_df)]

    def _get_output_filename(self, contrast: CorrelationContrast) -> str:
        """Build the filename from class attributes and provided contrast.

        Parameters
        ----------
        contrast : CorrelationContrast
            The contrast to use for building the filename.
        """
        return (
            f"group-{self.group_label}_correlation-{contrast.absolute_name}"
            f"-{contrast.sign}_measure-{self.feature_label}_fwhm-{self.fwhm}"
        )



from os import PathLike
from pathlib import Path
from typing import List, Optional, Union

import pandas as pd

__all__ = [
    "create_subs_sess_list",
    "write_list_of_files",
]


def create_subs_sess_list(
    input_dir: Union[str, PathLike],
    output_dir: Union[str, PathLike],
    file_name: Optional[str] = None,
    is_bids_dir: bool = True,
    use_session_tsv: bool = False,
):
    """Create the file subject_session_list.tsv that contains the list of the
    visits for each subject for a BIDS or CAPS compliant dataset.

    Parameters
    ----------
    input_dir : PathLike
        Path to the BIDS or CAPS directory.

    output_dir : PathLike
        Path to the output directory.

    file_name : str, optional
        The name of the output file.

    is_bids_dir : bool, optional
        Specify if input_dir is a BIDS directory or not (i.e. a CAPS directory).
        Default=True.

    use_session_tsv : bool
        Specify if the list uses the sessions listed in the sessions.tsv files.
        Default=False.
    """
    input_dir = Path(input_dir)
    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True)
    path_to_search = input_dir if is_bids_dir else input_dir / "subjects"
    txt = _create_subs_sess_list_as_text(path_to_search, use_session_tsv)
    file_name = file_name or "subjects_sessions_list.tsv"
    (output_dir / file_name).write_text(txt)


def _create_subs_sess_list_as_text(path_to_search: Path, use_session_tsv: bool) -> str:
    txt = "participant_id\tsession_id\n"
    subjects_paths = [f for f in path_to_search.glob("*sub-*")]
    subjects_paths.sort()
    if len(subjects_paths) == 0:
        raise IOError("Dataset empty or not BIDS/CAPS compliant.")
    for subject_path in subjects_paths:
        if use_session_tsv:
            txt += _create_session_list_for_subject_from_tsv(subject_path)
        else:
            sessions = [f for f in subject_path.glob("*ses-*")]
            sessions.sort()
            for ses_path in sorted(sessions):
                txt += f"{subject_path.name}\t{ses_path.name}\n"
    return txt


def _create_session_list_for_subject_from_tsv(subject_path: Path) -> str:
    if not (subject_path / f"{subject_path.name}_sessions.tsv").exists():
        raise ValueError(
            f"In dataset located at {subject_path.parent}, there is no session "
            f"TSV file for subject {subject_path.name}. Consider setting the "
            "argument `use_session_tsv` to False in order to rely on folder "
            "names parsing."
        )
    session_df = pd.read_csv(
        subject_path / f"{subject_path.name}_sessions.tsv", sep="\t"
    )
    session_df.dropna(how="all", inplace=True)
    session_list = sorted(list(session_df["session_id"].to_numpy()))
    return (
        "\n".join([f"{subject_path.name}\t{session}" for session in session_list])
        + "\n"
    )


def write_list_of_files(file_list: List[PathLike], output_file: PathLike) -> Path:
    """Save `file_list` list of files into `output_file` text file.

    Parameters
    ----------
    file_list : list of PathLike objects
        List of path to files to write.

    output_file : PathLike
        Path to the output txt file.

    Returns
    -------
    output_file : PathLike
        The path to the output file.

    Raises
    ------
    TypeError :
        If something else than a list was provided for file_list.
    IOError :
        If output_file already exists.
    """
    output_file = Path(output_file)
    if not isinstance(file_list, list):
        raise TypeError(
            f"`file_list` argument must be a list of paths. Instead {type(file_list)} was provided."
        )
    if output_file.is_file():
        raise IOError(f"Output file {output_file} already exists.")
    with open(output_file, "w") as fp:
        fp.write(f"\n".join([str(f) for f in file_list]))

    return output_file



"""Utility functions for the pipeline statistics volume.

Public functions are called within Nipype nodes in the pipeline.
These functions must be "self-contained". In other words, they should
have type hints composed of Python's builtins only, and they should
explicitly import everything they use (i.e. other functions).

Private functions are called by public functions of this module.
They can use more advanced abstractions.
"""

import functools
import operator
import typing as ty
from pathlib import Path


def get_group_1_and_2(tsv: str, contrast: str) -> tuple:
    """Based on the TSV file given in parameter, compute indexes of each group

    Parameters
    ----------
    tsv: str
        Path to the tsv file containing information on subjects/sessions with all covariates
    contrast: str
        Name of a column of the tsv

    Returns
    -------
    first_group_idx: list of int
        list of indexes of first group
    second_group_idx: list of int
        list of indexes of second group
    class_names: list of str of len 2
        list of the class names read in the column contrast of the tsv
    """
    import pandas as pds

    from clinica.utils.exceptions import ClinicaException

    # StatisticsVolume pipeline has been instantiated with tsv_file=tsv,
    # so check of existence and integrity have succeeded
    # No further check are done when trying to read it
    tsv = pds.read_csv(tsv, sep="\t")
    columns = list(tsv.columns)

    if contrast not in columns:
        raise ClinicaException(contrast + " is not present in " + tsv)

    class_names = sorted(list(set(tsv[contrast])))

    # This is a 2-sample t-test: we can only allow 2 classes
    if len(class_names) != 2:
        raise ClinicaException(
            "It must exist only 2 classes in the column "
            + contrast
            + " to perform 2-sample t-tests. Here Clinica found: "
            + str(class_names)
        )
    first_group_idx = [
        i for i, label in enumerate(list(tsv[contrast])) if label == class_names[0]
    ]
    second_group_idx = [
        i for i, label in enumerate(list(tsv[contrast])) if label == class_names[1]
    ]

    return first_group_idx, second_group_idx, class_names


def write_matlab_model(
    tsv: str,
    contrast: str,
    idx_group1: list,
    idx_group2: list,
    file_list: list,
    template_file: str,
):
    """Create the matlab .m file for the instantiation of the 2-sample t-test model in SPM

    Parameters
    ----------
    tsv: str
        Path to the tsv file containing information on subjects/sessions with all covariates
    contrast: str
        Name of a column of the tsv
    idx_group1: list of int
        List of indexes of first group
    idx_group2: list of int
        List of indexes of second group
    file_list: List
        List of files used in the statistical test. Their order is the same as it appears on the tsv file
    template_file: str
        Path to the template file used to generate the .m file

    Returns
    -------
    current_model: str
        Path to the matlab files with all the @TEXT replaced with the correct names
    covariates: list of str
        Names of covariates
    """
    from os import remove
    from os.path import abspath, isfile

    import pandas as pds

    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _convert_to_numeric,
        _create_spm_output_folder,
        _set_output_and_groups,
        _write_covariates,
    )
    from clinica.utils.exceptions import ClinicaException

    # Get template for model creation
    if not isfile(template_file):
        raise RuntimeError("[Error] " + template_file + " could not be found !")

    # Create current model filename
    current_model = abspath("./current_model_creation.m")

    if isfile(current_model):
        remove(current_model)

    with open(template_file, "r") as file:
        filedata = file.read()

    output_folder = _create_spm_output_folder(current_model)

    # Replace string in matlab file to set the output directory, and all the scans used for group 1 and 2
    _set_output_and_groups(
        output_folder,
        current_model,
        file_list,
        idx_group1,
        idx_group2,
        filedata,
    )

    # Add our covariates
    tsv = pds.read_csv(tsv, sep="\t")
    columns_stripped = [elem.strip(" ") for elem in list(tsv.columns)]
    if columns_stripped != list(tsv.columns):
        raise ClinicaException(
            "[Error] Check the column of your tsv file "
            + tsv
            + "Whitespace in the column names can cause errors"
        )
    covariates = [
        elem
        for elem in columns_stripped
        if elem not in ["participant_id", "session_id", contrast]
    ]
    for covar_number, covar in enumerate(covariates, start=1):
        current_covar_data = list(tsv[covar])
        if isinstance(current_covar_data[0], str):
            # Transform data
            current_covar_data = _convert_to_numeric(current_covar_data)
        _write_covariates(
            current_covar_data,
            idx_group1,
            idx_group2,
            current_model,
            covar,
            covar_number,
        )

    # Tell matlab to run the script at the end
    with open(current_model, "a") as file:
        file.write("spm_jobman('run', matlabbatch)")
    return current_model, covariates


def _create_spm_output_folder(current_model: str) -> str:
    """Creates the spm output folder.

    The folder is named '2_sample_t_test' and is located in the
    same folder as the provided model file.
    If the folder already exists, it is deleted and created again.

    Parameters
    ----------
    current_model: str
        Path to the matlab files with all the @TEXT replaced with the correct names

    Returns
    -------
    output_folder: str
        Path to the spm output_folder
    """
    from os import mkdir
    from os.path import abspath, dirname, isdir, join
    from shutil import rmtree

    output_folder = abspath(join(dirname(current_model), "..", "2_sample_t_test"))
    if isdir(output_folder):
        rmtree(output_folder)
    mkdir(output_folder)
    return output_folder


def _set_output_and_groups(
    output_folder: str,
    current_model: str,
    file_list: list,
    idx_group1: list,
    idx_group2: list,
    filedata: str,
) -> None:
    """Sets path to output and groups in the .m script

    Parameters
    ----------
    output_folder: str
        Path to the folder where the pipelines output will be generated
    current_model: str
        Path to the matlab files with all the @TEXT replaced with the correct names
    file_list: List
        List of files used in the statistical test. Their order is the same as it appears on the tsv file
    idx_group1: list of int
        List of indexes of first group
    idx_group2: list of int
        List of indexes of second group
    filedata: str
        .m script template which is to be modified and then run
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        _unravel_list_for_matlab,
    )

    filedata = filedata.replace("@OUTPUTDIR", "'" + output_folder + "'")
    filedata = filedata.replace(
        "@SCANS1",
        _unravel_list_for_matlab([file_list[i] for i in idx_group1]),
    )
    filedata = filedata.replace(
        "@SCANS2",
        _unravel_list_for_matlab([file_list[i] for i in idx_group2]),
    )
    with open(current_model, "w+") as file:
        file.write(filedata)


def _unravel_list_for_matlab(my_list: ty.List[str]) -> str:
    """Unpack a list into a Matlab compliant format to insert in .m files.

    Parameters
    ----------
    my_list: list of str

    Returns
    -------
    str
        Contains the different element of the list joined in one string.
    """
    result = "','".join(my_list)
    result = "'" + result + "'"
    return result


def _convert_to_numeric(data: ty.List[str]) -> ty.List[float]:
    """Convert the string values from the input list to numeric values.

    If the values can be casted to floats, then outputs a list of floats.
    If the values are categorical, then performs encoding of the categories.

    Parameters
    ----------
    data: list of str
        List of strings to be converted.

    Returns
    -------
    list :
        List of floats resulting from conversion.
    """
    import numpy as np

    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _is_number,
    )

    temp_data = [elem.replace(",", ".") for elem in data]
    if all(_is_number(elem) for elem in temp_data):
        return [float(elem) for elem in temp_data]
    # categorical variables (like Male; Female; M, F etc...)
    unique_values = list(np.unique(np.array(data)))
    return [unique_values.index(elem) for elem in data]


def _is_number(s: str) -> bool:
    """Returns True is the input can be converted to float, False otherwise."""
    try:
        # Try the conversion, if it is not possible, error will be raised
        float(s)
        return True
    except ValueError:
        return False
    except TypeError:
        return False


def _write_covariates(
    covariates: list,
    idx_group1: list,
    idx_group2: list,
    model: str,
    covar: int,
    covar_number: int,
) -> None:
    """Concatenate covariates for group1 and group2 and write them to disk.

    Parameters
    ----------
    covariates: list
        List of covariates to be concatenated and written.
    idx_group1: list of int
        List of indexes of first group
    idx_group2: list of int
        List of indexes of second group
    model: str
        Path to the matlab files with all the @TEXT replaced with the correct names
    covar: int
        Covariance
    covar_number: int
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _write_covariate_lines,
    )

    current_covar_data_group1 = [covariates[i] for i in idx_group1]
    current_covar_data_group2 = [covariates[i] for i in idx_group2]
    concatenated_covariates = current_covar_data_group1 + current_covar_data_group2
    _write_covariate_lines(model, covar_number, covar, concatenated_covariates)


def _write_covariate_lines(
    m_file_to_write_in: str,
    covar_number: int,
    covar_name: str,
    covar_values: list,
) -> None:
    """Add one line in the Matlab file `m_file_to_write_in` for each covariate.

    Parameters
    ----------
    m_file_to_write_in: str
        Path to the m-file
    covar_number: int
        This is the number of the covariate (must start at 1)
    covar_name: str
        Name of the covariate (ex: 'age', 'sex')
    covar_values: list of float
        Values of the covariates
    """
    from os.path import isfile

    if not isfile(m_file_to_write_in):
        raise FileNotFoundError(f"Could not find file {m_file_to_write_in}")
    covar_values_string = [str(elem) for elem in covar_values]
    covar_values_for_script = "[" + " ".join(covar_values_string) + "]"

    m_file = open(m_file_to_write_in, "a")
    m_file.write(
        "matlabbatch{1}.spm.stats.factorial_design.cov("
        + str(covar_number)
        + ").c = "
        + covar_values_for_script
        + ";\n"
    )
    m_file.write(
        "matlabbatch{1}.spm.stats.factorial_design.cov("
        + str(covar_number)
        + ").cname = '"
        + covar_name
        + "';\n"
    )
    m_file.write(
        "matlabbatch{1}.spm.stats.factorial_design.cov("
        + str(covar_number)
        + ").iCFI = 1;\n"
    )
    m_file.write(
        "matlabbatch{1}.spm.stats.factorial_design.cov("
        + str(covar_number)
        + ").iCC = 1;\n"
    )
    m_file.close()


def run_m_script(m_file: str) -> str:
    """Runs a Matlab file for SPM.

    Determines automatically if the script should be launched with SPM or SPM standalone.
    If launch with spm standalone, the line 'spm_jobman('run', matlabbatch)' must be removed because unnecessary

    Parameters
    ----------
    m_file: str
        Path to Matlab m file

    Returns
    -------
    output_mat_file: str
        Path to the SPM.mat file needed in SPM analysis
    """
    from pathlib import Path

    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _delete_last_line,
        _run_matlab_script_with_matlab,
        _run_matlab_script_with_spm_standalone,
    )
    from clinica.utils.spm import use_spm_standalone_if_available

    m_file = Path(m_file)
    if not m_file.exists():
        raise FileNotFoundError(f"[Error] File {m_file} does not exist")
    if m_file.suffix != ".m":
        raise ValueError(
            f"[Error] {m_file} is not a Matlab file (extension must be .m)"
        )
    if use_spm_standalone_if_available():
        _delete_last_line(m_file)
        _run_matlab_script_with_spm_standalone(m_file)
    else:
        _run_matlab_script_with_matlab(str(m_file))
    output_mat_file = (m_file.parent.parent / "2_sample_t_test" / "SPM.mat").resolve()
    if not output_mat_file.is_file():
        raise RuntimeError(f"Output matrix {output_mat_file} was not produced")
    return str(output_mat_file)


def _delete_last_line(filename: Path) -> None:
    """Removes the last line of the provided file.

    Used to remove the call to spm jobman if the MATLAB file
    is used with SPM standalone.

    Parameters
    ----------
    filename: Path
        Path to filename
    """
    import os

    with open(filename, "r+", encoding="utf-8") as file:
        # Move the pointer (similar to a cursor in a text editor) to the end of the file
        file.seek(0, os.SEEK_END)

        # This code means the following code skips the very last character in the file -
        # i.e. in the case the last line is null we delete the last line
        # and the penultimate one
        pos = file.tell() - 1

        # Read each character in the file one at a time from the penultimate
        # character going backwards, searching for a newline character
        # If we find a new line, exit the search
        while pos > 0 and file.read(1) != "\n":
            pos -= 1
            file.seek(pos, os.SEEK_SET)

        # So long as we're not at the start of the file, delete all the characters ahead
        # of this position
        if pos > 0:
            file.seek(pos, os.SEEK_SET)
            file.truncate()


def _run_matlab_script_with_spm_standalone(m_file: Path) -> None:
    """Runs a matlab script spm_standalone

    Parameters
    ----------
    m_file: str
        Path to the script
    """
    from os import system

    system(_get_matlab_standalone_command(m_file))


def _get_matlab_standalone_command(m_file: Path) -> str:
    """Get the matlab standalone command to be run depending on the user's OS.

    Notes
    -----
    On MAC systems, SPM standalone must be run directly from its root folder.

    Raises
    ------
    SystemError
        If this is run on unsupported platforms.
    """
    if _is_running_on_mac():
        return f"cd $SPMSTANDALONE_HOME && ./run_spm12.sh $MCR_HOME batch {m_file}"
    if _is_running_on_linux():
        return f"$SPMSTANDALONE_HOME/run_spm12.sh $MCR_HOME batch {m_file}"
    raise SystemError("Clinica only support Mac OS and Linux")


def _is_running_on_os(os: str) -> bool:
    import platform

    return platform.system().lower().startswith(os)


_is_running_on_mac = functools.partial(_is_running_on_os, os="darwin")
_is_running_on_linux = functools.partial(_is_running_on_os, os="linux")


def _run_matlab_script_with_matlab(m_file: str) -> None:
    """Runs a matlab script using matlab

    Parameters
    ----------
    m_file: str
        Path to the script
    """
    import platform
    from os.path import abspath

    from nipype.interfaces.matlab import MatlabCommand, get_matlab_command

    _add_spm_path_to_matlab_script(m_file)
    MatlabCommand.set_default_matlab_cmd(get_matlab_command())
    matlab = MatlabCommand()
    if platform.system().lower().startswith("linux"):
        matlab.inputs.args = "-nosoftwareopengl"
    matlab.inputs.paths = Path(m_file).parent
    matlab.inputs.script = Path(m_file).stem
    matlab.inputs.single_comp_thread = False
    matlab.inputs.logfile = abspath("./matlab_output.log")
    matlab.run()


def _add_spm_path_to_matlab_script(m_file: str) -> None:
    from clinica.utils.check_dependency import get_spm_home

    with open(m_file, "r") as fp:
        lines = fp.readlines()
    lines.insert(0, f"addpath('{get_spm_home()}');")
    with open(m_file, "w") as fp:
        fp.writelines(lines)


def clean_template_file(mat_file: str, template_file: str) -> str:
    """Make a copy of the template file (for estimation) and replace
    @SPMMAT by the real path to SPM.mat (mat_file).

    Parameters
    ----------
    mat_file: str
        Path to the SPM.mat file of the SPM analysis
    template_file: str
        Path to the template file for the estimation of the model

    Returns
    -------
    current_model_estimation: str
        Path to the template file filled with SPM.mat information, ready to be launched
    """
    from os.path import abspath

    with open(template_file, "r") as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace("@SPMMAT", "'" + mat_file + "'")
    current_model_estimation = abspath("./current_model_creation.m")
    with open(current_model_estimation, "w+") as file:
        file.write(filedata)

    return current_model_estimation


def clean_spm_result_file(
    mat_file: str,
    template_file: str,
    method: str,
    threshold: float,
) -> str:
    """Make a copy of the template file (for results) and replace
    @SPMMAT by the real path to SPM.mat (mat_file).

    Parameters
    ----------
    mat_file: str
        Path to the SPM.mat file of the SPM analysis
    template_file: str
        Path to the template file for getting the results of the model
    method: str
        method. In our case, "none"
    threshold: float
        cluster threshold

    Returns
    -------
    current_model_estimation: str
        Path to the template file filled with SPM.mat information, ready to be launched
    """
    import time
    from os.path import abspath

    # Read template
    with open(template_file, "r") as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace("@SPMMAT", "'" + mat_file + "'")
    filedata = filedata.replace("@CORRECTIONMETHOD", "'" + method + "'")
    filedata = filedata.replace("@THRESH", str(threshold))
    if method != "none":
        # Necessary so that the 2 computations does not overlap
        time.sleep(15)
    current_model_result = abspath("./current_model_result.m")
    with open(current_model_result, "w+") as file:
        file.write(filedata)

    return current_model_result


def clean_spm_contrast_file(
    mat_file: str,
    template_file: str,
    covariates: list,
    class_names: list,
):
    """Make a copy of the template file (for results) and replace
    @SPMMAT, @COVARNUMBER, @GROUP1, @GROUP2 by the corresponding variables.

    Parameters
    ----------
    mat_file: str
        Path to the SPM.mat file of the SPM analysis
    template_file: str
        Path to the template file for getting the results of the model
    covariates: list of str
        List of covariates
    class_names: list of str
        Corresponds to the 2 classes for the group comparison

    Returns
    -------
    current_model_estimation: str
        Path to the template file filled with variables, ready to be launched
    """
    from os.path import abspath

    number_of_covariates = len(covariates)

    with open(template_file, "r") as file:
        filedata = file.read()
    # Replace by the real path to spm.mat
    filedata = filedata.replace("@SPMMAT", "'" + mat_file + "'")
    filedata = filedata.replace("@COVARNUMBER", "0 " * number_of_covariates)
    filedata = filedata.replace("@GROUP1", "'" + class_names[0] + "'")
    filedata = filedata.replace("@GROUP2", "'" + class_names[1] + "'")
    current_model_estimation = abspath("./current_model_contrast.m")
    with open(current_model_estimation, "w+") as file:
        file.write(filedata)

    return current_model_estimation


def copy_and_rename_spm_output_files(
    spm_mat: str,
    class_names: list,
    covariates: list,
    group_label: str,
    fwhm: int,
    measure: str,
    output_dir: str = ".",
) -> tuple:
    """Once analysis is done, copy and rename (according to class names)
    the different filenames in the provided output directory.

    Parameters
    ----------
    spm_mat: str
        Path to the SPM.mat file of the SPM analysis

    class_names: list of str
        Corresponds to the 2 classes for the group comparison

    covariates: list of str
        List of covariates.

    group_label: str
        Name of the group label.

    fwhm: int
        Fwhm in mm used.

    measure: str
        Measure used.

    output_dir : str, optional
        Output folder where the copied figures should be written.
        Default to current folder.

    Returns
    -------
    spm_T_maps : list of str
        List of two SPM T maps for first and second group comparison.

    spm_figures : list of str
        Path to figure files.

    other_spm_files : list of str
        List of three files produced by SPM:
            - variance of error
            - resels per voxels
            - mask of included voxels

    regression_coeff : list of str
        Path to regression coefficients.

    contrasts : list of str
        Path to weighted parameter estimation for the 2 contrasts.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _check_spm_and_output_dir,
        _rename_beta_files,
        _rename_other_spm_files,
        _rename_spm_contrast_files,
        _rename_spm_figures,
        _rename_spm_t_maps,
    )

    spm_dir, output_dir = _check_spm_and_output_dir(spm_mat, output_dir)
    parameters = {
        "group_label": group_label,
        "measure": measure,
        "class_names": class_names,
        "fwhm": fwhm,
        "covariates": covariates,
    }
    rename_functions = (
        _rename_spm_t_maps,
        _rename_spm_figures,
        _rename_other_spm_files,
        _rename_beta_files,
        _rename_spm_contrast_files,
    )
    return tuple(
        [
            func(
                spm_dir=spm_dir,
                output_dir=output_dir,
                parameters_for_new_filename_construction=parameters,
            )
            for func in rename_functions
        ]
    )


def _rename_spm_t_maps(
    spm_dir: Path,
    output_dir: Path,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _build_t_map_filenames,
        _get_spm_t_maps,
    )

    return _rename_spm_files(
        spm_dir,
        output_dir,
        spm_filename_getter=_get_spm_t_maps,
        new_filename_constructor=_build_t_map_filenames,
        parameters_for_new_filename_construction=parameters_for_new_filename_construction,
    )


def _rename_spm_files(
    spm_dir: Path,
    output_dir: Path,
    spm_filename_getter: ty.Callable,
    new_filename_constructor: ty.Callable,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    """Rename SPM files in spm_dir to new file names in output_dir.

    The function consists of three steps:
        - Get the desired spm files in spm_dir.
          This is achieved by the provided spm_filename_getter function.
        - Build the new filenames from provided parameters.
          This is achieved by the provided new_filename_constructor function.
        - Copy the files from spm_dir to output_dir and perform the renaming
          at the same time.

    Parameters
    ----------
    spm_dir : Path
        The path to the SPM folder where SPM output files are located.

    output_dir : Path
        The path to the desired output folder where the renamed spm files
        should be written.

    spm_filename_getter : Callable
        The function responsible for getting the desired SPM files.
        These functions expect the spm_dir as input and should be
        partial implementations of the function `_get_spm_files`.

    new_filename_constructor : Callable
        The function responsible for building the new file names.
        It is expected that the spm_filename_getter and the
        new_filename_constructor provide old and new names in the
        correct order.

    parameters_for_new_filename_construction : dict
        The dictionary of parameters which will be passed to the
        new_filename_constructor for building the new file names.
        This consists of information relative to the class names,
        the group label, the measure used, the smoothing kernel,
        and the covariates.

    Returns
    -------
    list of str :
        The list of new file names.
    """
    kwargs = dict()
    # The beta file getter needs to know the number of covariates
    # to verify that it has the correct number of files.
    if spm_filename_getter.__name__ == "_get_spm_beta_files":
        kwargs["expected_number_of_files"] = 2 + len(
            parameters_for_new_filename_construction["covariates"]
        )
    existing_spm_filenames = spm_filename_getter(spm_dir, **kwargs)
    new_filenames = new_filename_constructor(
        **parameters_for_new_filename_construction,
        existing_spm_filenames=existing_spm_filenames,
    )
    spm_to_clinica_filename_mapping = {
        k: v for k, v in zip(existing_spm_filenames, new_filenames)
    }
    _copy_spm_to_clinica_files(spm_dir, output_dir, spm_to_clinica_filename_mapping)

    return [str(output_dir / f) for f in new_filenames]


def _copy_spm_to_clinica_files(
    spm_dir: Path, output_dir: Path, name_mapping: ty.Dict[str, str]
) -> None:
    """Copy files in spm_dir to output_dir while renaming them.

    The renaming is controlled by the 'old_name:new_name' mapping 'name_mapping'.

    Parameters
    ----------
    spm_dir : Path
        The path to the SPM folder where SPM output files are located.

    output_dir : Path
        The path to the desired output folder where the renamed spm files
        should be written.

    name_mapping : dict
        Dictionary implementing the mapping between old filenames (the ones
        in the spm_dir) and desired new file names (the ones that will be
        written in output_dir).
    """
    _copy_files(
        {str(spm_dir / k): str(output_dir / v) for k, v in name_mapping.items()}
    )


def _copy_files(source_to_destination_mapping: ty.Dict[str, str]) -> None:
    """Copy the source files in the mapping keys to the destinations in the values.

    Parameters
    ----------
    source_to_destination_mapping : dict
        The mapping between old file paths and new file paths.
    """
    from shutil import copyfile

    for source, destination in source_to_destination_mapping.items():
        copyfile(source, destination)


def _rename_spm_contrast_files(
    spm_dir: Path,
    output_dir: Path,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    """Rename the SPM contrast files."""
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _get_new_spm_contrast_files,
        _get_spm_contrast_files,
    )

    return _rename_spm_files(
        spm_dir,
        output_dir,
        spm_filename_getter=_get_spm_contrast_files,
        new_filename_constructor=_get_new_spm_contrast_files,
        parameters_for_new_filename_construction=parameters_for_new_filename_construction,
    )


def _rename_beta_files(
    spm_dir: Path,
    output_dir: Path,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _get_new_spm_beta_files,
        _get_spm_beta_files,
    )

    return _rename_spm_files(
        spm_dir,
        output_dir,
        spm_filename_getter=_get_spm_beta_files,
        new_filename_constructor=_get_new_spm_beta_files,
        parameters_for_new_filename_construction=parameters_for_new_filename_construction,
    )


def _rename_spm_figures(
    spm_dir: Path,
    output_dir: Path,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    """Rename the SPM figure files."""
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _get_new_spm_figures,
        _get_spm_figures,
    )

    return _rename_spm_files(
        spm_dir,
        output_dir,
        spm_filename_getter=_get_spm_figures,
        new_filename_constructor=_get_new_spm_figures,
        parameters_for_new_filename_construction=parameters_for_new_filename_construction,
    )


def _rename_other_spm_files(
    spm_dir: Path,
    output_dir: Path,
    parameters_for_new_filename_construction: dict,
) -> ty.List[str]:
    """Rename other SPM files of interest."""
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (  # noqa
        _get_new_other_spm_files,
        _get_other_spm_files,
    )

    return _rename_spm_files(
        spm_dir,
        output_dir,
        spm_filename_getter=_get_other_spm_files,
        new_filename_constructor=_get_new_other_spm_files,
        parameters_for_new_filename_construction=parameters_for_new_filename_construction,
    )


def _get_spm_t_maps(spm_dir: Path) -> ty.List[str]:
    return _get_spm_files(spm_dir, pattern="spmT", file_type="t-map(s)")


def _get_other_spm_files(spm_dir: Path, **kwargs) -> ty.List[str]:
    return ["ResMS.nii", "RPV.nii", "mask.nii"]


def _get_spm_figures(spm_dir: Path) -> ty.List[str]:
    import operator

    return _get_spm_files(
        spm_dir,
        pattern="png",
        grab_by_prefix=False,
        file_type="figures",
        op=operator.ge,
    )


def _get_spm_beta_files(spm_dir: Path, expected_number_of_files: int) -> ty.List[str]:
    return _get_spm_files(
        spm_dir,
        pattern="beta_",
        expected_number_of_files=expected_number_of_files,
        file_type="betas",
    )


def _get_spm_contrast_files(spm_dir: Path) -> ty.List[str]:
    return _get_spm_files(spm_dir, pattern="con_", file_type="contrast")


def _get_spm_files(
    spm_dir: Path,
    pattern: str,
    file_type: str,
    grab_by_prefix: bool = True,
    expected_number_of_files: int = 2,
    op=operator.eq,
) -> ty.List[str]:
    """Get the SPM file names of interest in spm_dir.

    Parameters
    ----------
    spm_dir : Path
        The path to the SPM folder where the SPM files are located.

    pattern : str
        Pattern to look for in the file names. This can be searched as
        a prefix or a suffix (controlled by 'grab_by_prefix').

    file_type : str
        The kind of file that the function will be looking for.
        This is only used for error messages and should be informative
        for end users.

    grab_by_prefix : bool, optional
        If True, look for pattern as a prefix in the file names.
        Otherwise, look for pattern as a suffix in the file names.
        Default=True.

    expected_number_of_files : int, optional
        The number of files expected in the spm_dir if the SPM processing
        took place as expected. If the number of files found does not satisfy
        the constraint, a RunTimeError will be raised.

    op : operator, optional
        Used to compare the number of files found to the expected number of files.
        This is set to equality by default such that an error will be raised if
        the number of found files is not equal to the desired number.
        Some functions expect at least some number of files such that 'op.ge'
        can be used instead.

    Returns
    -------
    list of str :
        List of file names found in spm_dir that match the criteria.
    """
    spm_files = [
        f.name
        for f in spm_dir.iterdir()
        if _is_valid_filename(f.name, pattern, grab_by_prefix)
    ]
    if not op(len(spm_files), expected_number_of_files):
        moderator = " at least" if op == operator.ge else ""
        raise RuntimeError(
            f"[Error] Wrong number of SPM {file_type} files. "
            f"Expected{moderator} {expected_number_of_files}, got {len(spm_files)}."
        )

    return sorted(spm_files)


def _is_valid_filename(filename: str, pattern: str, grab_by_prefix: bool) -> bool:
    """Filename is valid if it doesn't start with '.' and if it has the provided
    pattern as a prefix or as a suffix.
    """
    pattern_condition = (
        filename.startswith(pattern) if grab_by_prefix else filename.endswith(pattern)
    )

    return not filename.startswith(".") and pattern_condition


def _get_new_other_spm_files(group_label: str, **kwargs):
    return [
        f"group-{group_label}_VarianceError.nii",
        "resels_per_voxel.nii",
        "included_voxel_mask.nii",
    ]


def _get_new_spm_figures(
    group_label: str, existing_spm_filenames: ty.List[str], **kwargs
) -> ty.List[str]:
    fig_number = [
        int(f[-7:-4]) for f in existing_spm_filenames
    ]  # assumes number is encoded on 3 digits
    return [f"group-{group_label}_report-{i}.png" for i in fig_number]


def _check_spm_and_output_dir(
    spm_mat: str, output_dir: ty.Optional[str] = None
) -> ty.Tuple[Path, Path]:
    from pathlib import Path

    spm_mat = Path(spm_mat)
    spm_dir = spm_mat.parent

    if not spm_mat.is_file():
        if not spm_dir.is_dir():
            raise RuntimeError(f"[Error] output folder {spm_dir} does not exist.")
        raise RuntimeError(f"[Error] SPM matrix {spm_mat} does not exist.")

    output_dir = Path(output_dir) if output_dir else Path.cwd()

    return spm_dir.resolve(), output_dir.resolve()


def _build_t_map_filenames(
    class_names: ty.List[str],
    group_label: str,
    measure: str,
    fwhm: int,
    **kwargs,
) -> ty.List[str]:
    """Build the T-map filenames from the class names, group label, measure, and fwhm."""
    fwhm_string = f"_fwhm-{int(fwhm)}" if fwhm else ""

    return [
        f"group-{group_label}_{contrast}_measure-{measure}{fwhm_string}_TStatistics.nii"
        for contrast in _build_contrasts_from_class_names(class_names)
    ]


def _build_contrasts_from_class_names(class_names: ty.List[str]) -> ty.List[str]:
    """Build the contrast strings from the class names."""
    return [
        f"{class_names[0]}-lt-{class_names[1]}",
        f"{class_names[1]}-lt-{class_names[0]}",
    ]


def _get_new_spm_beta_files(
    class_names: ty.List[str], covariates: ty.List[str], **kwargs
) -> ty.List[str]:
    return [f"{c}.nii" for c in class_names] + [f"{covar}.nii" for covar in covariates]


def _get_new_spm_contrast_files(
    group_label: str, class_names: ty.List[str], measure: str, **kwargs
) -> ty.List[str]:
    return [
        f"group-{group_label}_{contrast}_measure-{measure}_contrast.nii"
        for contrast in _build_contrasts_from_class_names(class_names)
    ]



import os
import typing as ty

import pydra


def input_workflow(name: str = "input_workflow", **inputs) -> pydra.Workflow:
    from pydra.tasks import bids

    workflow = pydra.Workflow(
        name=name,
        input_spec=["bids_dir"],
        **inputs,
    )

    workflow.add(
        bids.read_bids_dataset(
            output_queries={
                "out": {
                    "suffix": "T1w",
                    "extension": ["nii", "nii.gz"],
                }
            },
            dataset_path=workflow.lzin.bids_dir,
            name="read_t1w",
        )
    )

    workflow.set_output([("t1w", workflow.read_t1w.lzout.out)])

    return workflow


def core_workflow(name: str = "core_workflow", **inputs) -> pydra.Workflow:
    from pydra.tasks import bids, freesurfer

    from .tasks import compute_freesurfer_subject_id

    workflow = pydra.Workflow(
        name=name,
        input_spec=["caps_dir", "t1w"],
        **inputs,
    )

    workflow.add(
        bids.parse_bids_name(
            output_entities={
                "participant_id": "sub",
                "session_id": "ses",
            },
            file_path=workflow.lzin.t1w,
            name="parse_t1w",
        )
    )

    workflow.add(
        compute_freesurfer_subject_id(
            participant_id=workflow.parse_t1w.lzout.participant_id,
            session_id=workflow.parse_t1w.lzout.session_id,
            name="compute_freesurfer_subject_id",
        )
    )

    workflow.add(
        freesurfer.ReconAll(
            name="recon_all",
            t1_volume_file=workflow.lzin.t1w,
            subject_id=workflow.compute_freesurfer_subject_id.lzout.subject_id,
            subjects_dir=workflow.lzin.caps_dir,
        )
    )

    workflow.set_output(
        [
            ("subjects_dir", workflow.recon_all.lzout.subjects_dir),
            ("subject_id", workflow.recon_all.lzout.subject_id),
        ]
    )

    return workflow


def build_workflow(
    bids_dir: os.PathLike,
    caps_dir: os.PathLike,
    atlas_path: os.PathLike,
    name: str = "t1_freesurfer",
) -> pydra.Workflow:
    workflow = pydra.Workflow(
        name=name,
        input_spec=["bids_dir", "caps_dir"],
        bids_dir=bids_dir,
    )

    workflow.add(
        input_workflow(
            name="input_workflow",
            bids_dir=workflow.lzin.bids_dir,
        )
    )

    workflow.add(
        core_workflow(
            name="core_workflow",
            caps_dir=caps_dir,
            t1w=workflow.input_workflow.lzout.t1w,
        ).split("t1w")
    )

    workflow.set_output(
        [
            ("t1w", workflow.input_workflow.lzout.t1w),
            ("subjects_dir", workflow.core_workflow.lzout.subjects_dir),
            ("subject_id", workflow.core_workflow.lzout.subject_id),
        ]
    )

    return workflow



def run_n4biasfieldcorrection_task(
    input_image: str,
    bspline_fitting_distance: int,
    output_prefix=None,
    output_dir=None,
    save_bias=False,
    verbose=False,
) -> str:
    from pathlib import Path

    from clinica.pipelines.t1_linear.anat_linear_utils import run_n4biasfieldcorrection

    if output_dir:
        output_dir = Path(output_dir)

    return str(
        run_n4biasfieldcorrection(
            Path(input_image),
            bspline_fitting_distance,
            output_prefix,
            output_dir,
            save_bias,
            verbose,
        )
    )


def run_ants_registration_task(
    fixed_image: str,
    moving_image: str,
    random_seed: int,
    output_prefix=None,
    output_dir=None,
) -> tuple:
    from pathlib import Path

    from clinica.pipelines.t1_linear.anat_linear_utils import run_ants_registration

    if output_dir:
        output_dir = Path(output_dir)

    warped_image_output_path, transformation_matrix_output_path = run_ants_registration(
        Path(fixed_image),
        Path(moving_image),
        random_seed,
        output_prefix,
        output_dir,
    )

    return str(warped_image_output_path), str(transformation_matrix_output_path)



from typing import List, Optional, Tuple

import click

import clinica.pydra.engine_utils as pydra_utils
import clinica.pydra.pet_linear.pipeline as pydra_pet_linear
from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "pydra-pet-linear"


@clinica_pipeline
@click.command(name=pipeline_name, hidden=True)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.argument.acq_label
@cli_param.argument.suvr_reference_region
@cli_param.option_group.pipeline_specific_options
@cli_param.option_group.option(
    "-ui",
    "--uncropped_image",
    is_flag=True,
    help="Do not crop the image with template (cropped image are suggested for using with DL models)",
)
@cli_param.option_group.option(
    "--save_pet_in_t1w_space",
    is_flag=True,
    help="Save the PET image in the T1w space computed in the intermediate step of the pipeline",
)
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
def cli(
    bids_directory: str,
    caps_directory: str,
    acq_label: str,
    suvr_reference_region: str,
    uncropped_image: bool = False,
    save_pet_in_t1w_space: bool = False,
    n_procs: Optional[int] = None,
) -> None:
    """Affine registration of PET images to the MNI standard space (Pydra engine).

    Prerequisite: You need to have performed the t1-linear pipeline on your T1-weighted MR images.
    See https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/PET_Linear/"

    Parameters
    ----------
    bids_directory : str
        Path to the input folder containing a BIDS compliant dataset.

    caps_directory : str
        Path to a CAPS folder containing the results of the required pipelines (T1-linear).
        This will also be used for writing the outputs.

    acq_label : str
        Label given to the PET acquisition, specifying the tracer used.
        Frequently used values are '18FFDG' or '18FAV45'.

    suvr_reference_region : {'pons', 'cerebellumPons', 'pons2', 'cerebellumPons2'}
        Reference region used for intensity normalization.

    uncropped_image : bool, optional
        Whether to also return the uncropped image or not. Default=False.

    save_pet_in_t1w_space : bool, optional
        Save PET imaging data registered to corresponding T1 space. Default=False.

    n_procs : int, optional
        Number of processes to use.
    """
    parameters = {
        "acq_label": acq_label,
        "suvr_reference_region": suvr_reference_region,
        "uncropped_image": uncropped_image,
        "save_PETinT1w": save_pet_in_t1w_space,
    }
    pipeline = pydra_pet_linear.build_core_workflow(
        name="pet-linear-pydra",
        input_dir=bids_directory,
        output_dir=caps_directory,
        parameters=parameters,
    )
    pydra_utils.run(pipeline, n_procs=n_procs)


if __name__ == "__main__":
    cli()



from typing import Optional

import click

from clinica import option
from clinica.pipelines import cli_param
from clinica.pipelines.engine import clinica_pipeline

pipeline_name = "dwi-preprocessing-using-t1"


@clinica_pipeline
@click.command(name=pipeline_name)
@cli_param.argument.bids_directory
@cli_param.argument.caps_directory
@cli_param.option_group.pipeline_specific_options
@cli_param.option.low_bval
@cli_param.option_group.common_pipelines_options
@cli_param.option.subjects_sessions_tsv
@cli_param.option.working_directory
@option.global_option_group
@option.n_procs
@cli_param.option_group.advanced_pipeline_options
@cli_param.option.use_cuda
@cli_param.option.initrand
@cli_param.option.delete_cache
@cli_param.option.random_seed
@cli_param.option.caps_name
def cli(
    bids_directory: str,
    caps_directory: str,
    low_bval: int = 5,
    subjects_sessions_tsv: Optional[str] = None,
    working_directory: Optional[str] = None,
    n_procs: Optional[int] = None,
    use_cuda: bool = False,
    initrand: bool = False,
    delete_cache: bool = False,
    random_seed: Optional[int] = None,
    caps_name: Optional[str] = None,
) -> None:
    """Preprocessing of raw DWI datasets using a T1w image.

    https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/DWI_Preprocessing/
    """
    from networkx import Graph

    from clinica.utils.ux import print_end_pipeline

    from .pipeline import DwiPreprocessingUsingT1

    parameters = {
        "low_bval": low_bval,
        "use_cuda": use_cuda,
        "initrand": initrand,
        "delete_cache": delete_cache,
        "random_seed": random_seed,
        "double_precision": True,  # TODO: expose to user ?
    }

    pipeline = DwiPreprocessingUsingT1(
        bids_directory=bids_directory,
        caps_directory=caps_directory,
        tsv_file=subjects_sessions_tsv,
        base_dir=working_directory,
        parameters=parameters,
        name=pipeline_name,
        caps_name=caps_name,
    )

    exec_pipeline = (
        pipeline.run(plugin="MultiProc", plugin_args={"n_procs": n_procs})
        if n_procs
        else pipeline.run()
    )

    if isinstance(exec_pipeline, Graph):
        print_end_pipeline(
            pipeline_name, pipeline.base_dir, pipeline.base_dir_was_specified
        )


if __name__ == "__main__":
    cli()



import typing as ty

from pydra.mark import annotate, task


@task
@annotate({"return": {"idx_group1": list, "idx_group2": list, "class_names": list}})
def get_group_1_and_2_task(tsv: str, contrast: str) -> dict:
    """Pydra task for computing indexes of each group.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.get_group_1_and_2`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        get_group_1_and_2,
    )

    return get_group_1_and_2(tsv, contrast)


@task
@annotate({"return": {"current_model": str, "covariates": list}})
def write_matlab_model_task(
    tsv: str,
    contrast: str,
    idx_group1: list,
    idx_group2: list,
    file_list: list,
    template_file: str,
):
    """Pydra task for creating the .m file for the instantiation of the 2-sample t-test model in SPM

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.model_creation`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        write_matlab_model,
    )

    return write_matlab_model(
        tsv, contrast, idx_group1, idx_group2, file_list, template_file
    )


@task
@annotate({"return": {"output_mat_file": str}})
def run_m_script_task(m_file: str) -> str:
    """Pydra task for running an .m script

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.run_m_script`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import run_m_script

    return run_m_script(m_file)


@task
@annotate({"return": {"current_model_estimation": str}})
def clean_template_file_task(mat_file: str, template_file: str) -> str:
    """Pydra task for estimation script.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.estimate`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        clean_template_file,
    )

    return clean_template_file(mat_file, template_file)


@task
@annotate({"return": {"current_model_result": str}})
def clean_spm_result_file_task(
    mat_file: str, template_file: str, method: str, threshold: float
) -> str:
    """Pydra task for result script.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.results`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        clean_spm_result_file,
    )

    return clean_spm_result_file(mat_file, template_file, method, threshold)


@task
@annotate({"return": {"current_model_estimation": str}})
def clean_spm_contrast_file_task(
    mat_file: str, template_file: str, covariates: list, class_names: list
):
    """Pydra task for contrast script.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.contrast`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        clean_spm_contrast_file,
    )

    return clean_spm_contrast_file(mat_file, template_file, covariates, class_names)


@task
@annotate(
    {
        "return": {
            "spmT_0001": str,
            "spmT_0002": str,
            "spm_figures": list,
            "variance_of_error": str,
            "resels_per_voxels": str,
            "mask": str,
            "regression_coeff": list,
            "contrasts": list[str],
        }
    }
)
def copy_and_rename_spm_output_files_task(
    spm_mat: str,
    class_names: list,
    covariates: list,
    group_label: str,
    fwhm: int,
    measure: str,
):
    """Pydra task for reading outputs of the statistics volume pipeline.

    ..note::
        Please refer to the documentation of function
        `clinica.pipelines.statistics_volume.statistics_volume_utils.read_output`.
    """
    from clinica.pipelines.statistics_volume.statistics_volume_utils import (
        copy_and_rename_spm_output_files,
    )

    return copy_and_rename_spm_output_files(
        spm_mat, class_names, covariates, group_label, fwhm, measure
    )
